"""Score visualization interface.

Unicode encoding for musical symbols is based on Standard Music Font
Layout (SMuFL) (see https://w3c.github.io/smufl/gitbook/).

"""
import warnings
from operator import attrgetter
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.font_manager import FontProperties
from matplotlib.lines import Line2D
from matplotlib.patches import Arc, Rectangle
from matplotlib.text import Text

from ..base import Base
from ..classes import KeySignature, Note, Tempo, TimeSignature
from ..external import get_bravura_font_path

if TYPE_CHECKING:
    from ..music import Music


COMMON_NOTE_CODES = {
    2: "\uE1D0",  # double whole note
    1: "\uE1D2",  # whole note
    0.5: "\uE1D3",  # half note
    0.25: "\uE1D5",  # quarter note
    0.125: "\uE1D7",  # 8th note
    0.0625: "\uE1D9",  # 16th note
    0.03125: "\uE1DB",  # 32th note
    0.015625: "\uE1DD",  # 64th note
    0.0078125: "\uE1DF",  # 128th note
    0.00390625: "\uE1E1",  # 256th note
    0.001953125: "\uE1E3",  # 512th note
    0.0009765625: "\uE1E5",  # 1024th note
    # Dotted
    3: "\uE1D0 \uE1E7",
    1.5: "\uE1D2 \uE1E7",
    0.75: "\uE1D3 \uE1E7",
    0.375: "\uE1D5 \uE1E7",
    0.1875: "\uE1D7 \uE1E7",
    0.09375: "\uE1D9 \uE1E7",
    0.046875: "\uE1DB \uE1E7",
    0.0234375: "\uE1DD \uE1E7",
    0.01171875: "\uE1DF \uE1E7",
    0.005859375: "\uE1E1 \uE1E7",
    0.0029296875: "\uE1E3 \uE1E7",
    0.00146484375: "\uE1E5 \uE1E7",
    # Double dotted
    3.5: "\uE1D0 \uE1E7 \uE1E7",
    1.75: "\uE1D2 \uE1E7 \uE1E7",
    0.875: "\uE1D3 \uE1E7 \uE1E7",
    0.4375: "\uE1D5 \uE1E7 \uE1E7",
    0.21875: "\uE1D7 \uE1E7 \uE1E7",
    0.109375: "\uE1D9 \uE1E7 \uE1E7",
    0.0546875: "\uE1DB \uE1E7 \uE1E7",
    0.02734375: "\uE1DD \uE1E7 \uE1E7",
    0.013671875: "\uE1DF \uE1E7 \uE1E7",
    0.0068359375: "\uE1E1 \uE1E7 \uE1E7",
    0.00341796875: "\uE1E3 \uE1E7 \uE1E7",
    0.001708984375: "\uE1E5 \uE1E7 \uE1E7",
}


COMMON_NOTE_CODES_ALT = {
    2: "\uE1D0",  # double whole note
    1: "\uE1D2",  # whole note
    0.5: "\uE1D3",  # half note
    0.25: "\uE1D5",  # quarter note
    0.125: "\uE1D5",  # 8th note
    0.0625: "\uE1D5",  # 16th note
    0.03125: "\uE1D5",  # 32th note
    0.015625: "\uE1D5",  # 64th note
    0.0078125: "\uE1D5",  # 128th note
    0.00390625: "\uE1D5",  # 256th note
    0.001953125: "\uE1D5",  # 512th note
    0.0009765625: "\uE1D5",  # 1024th note
    # Dotted
    3: "\uE1D0 \uE1E7",
    1.5: "\uE1D2 \uE1E7",
    0.75: "\uE1D3 \uE1E7",
    0.375: "\uE1D5 \uE1E7",
    0.1875: "\uE1D5 \uE1E7",
    0.09375: "\uE1D5 \uE1E7",
    0.046875: "\uE1D5 \uE1E7",
    0.0234375: "\uE1D5 \uE1E7",
    0.01171875: "\uE1D5 \uE1E7",
    0.005859375: "\uE1D5 \uE1E7",
    0.0029296875: "\uE1D5 \uE1E7",
    0.00146484375: "\uE1D5 \uE1E7",
    # Double dotted
    3.5: "\uE1D0 \uE1E7 \uE1E7",
    1.75: "\uE1D2 \uE1E7 \uE1E7",
    0.875: "\uE1D3 \uE1E7 \uE1E7",
    0.4375: "\uE1D5 \uE1E7 \uE1E7",
    0.21875: "\uE1D5 \uE1E7 \uE1E7",
    0.109375: "\uE1D5 \uE1E7 \uE1E7",
    0.0546875: "\uE1D5 \uE1E7 \uE1E7",
    0.02734375: "\uE1D5 \uE1E7 \uE1E7",
    0.013671875: "\uE1D5 \uE1E7 \uE1E7",
    0.0068359375: "\uE1D5 \uE1E7 \uE1E7",
    0.00341796875: "\uE1D5 \uE1E7 \uE1E7",
    0.001708984375: "\uE1D5 \uE1E7 \uE1E7",
}

COMMON_REST_CODES = {
    2: "\uE4E2",  # double whole rest
    1: "\uE4E3",  # whole rest
    0.5: "\uE1E4",  # half rest
    0.25: "\uE1E5",  # quarter rest
    0.125: "\uE1E6",  # 8th rest
    0.0625: "\uE1E7",  # 16th rest
    0.03125: "\uE1E8",  # 32th rest
    0.015625: "\uE1E9",  # 64th rest
    0.0078125: "\uE1EA",  # 128th rest
    0.00390625: "\uE1EB",  # 256th rest
    0.001953125: "\uE1EC",  # 512th rest
    0.0009765625: "\uE1ED",  # 1024th rest
    # Dotted
    3: "\uE4E2 \uE1E7",
    1.5: "\uE4E3 \uE1E7",
    0.75: "\uE1E4 \uE1E7",
    0.375: "\uE1E5 \uE1E7",
    0.1875: "\uE1E6 \uE1E7",
    0.09375: "\uE1E7 \uE1E7",
    0.046875: "\uE1E8 \uE1E7",
    0.0234375: "\uE1E9 \uE1E7",
    0.01171875: "\uE1EA \uE1E7",
    0.005859375: "\uE1EB \uE1E7",
    0.0029296875: "\uE1EC \uE1E7",
    0.00146484375: "\uE1ED \uE1E7",
    # Double dotted
    3.5: "\uE4E2 \uE1E7 \uE1E7",
    1.75: "\uE4E3 \uE1E7 \uE1E7",
    0.875: "\uE1E4 \uE1E7 \uE1E7",
    0.4375: "\uE1E5 \uE1E7 \uE1E7",
    0.21875: "\uE1E6 \uE1E7 \uE1E7",
    0.109375: "\uE1E7 \uE1E7 \uE1E7",
    0.0546875: "\uE1E8 \uE1E7 \uE1E7",
    0.02734375: "\uE1E9 \uE1E7 \uE1E7",
    0.013671875: "\uE1EA \uE1E7 \uE1E7",
    0.0068359375: "\uE1EB \uE1E7 \uE1E7",
    0.00341796875: "\uE1EC \uE1E7 \uE1E7",
    0.001708984375: "\uE1ED \uE1E7 \uE1E7",
}

NOTE_CODES = [
    "\uE1D0",  # double whole note
    "\uE1D2",  # whole note
    "\uE1D3",  # half note
    "\uE1D5",  # quarter note
    "\uE1D7",  # 8th note
    "\uE1D9",  # 16th note
    "\uE1DB",  # 32th note
    "\uE1DD",  # 64th note
    "\uE1DF",  # 128th note
    "\uE1E1",  # 256th note
    "\uE1E3",  # 512th note
    "\uE1E5",  # 1024th note
]

NOTE_CODES_ALT = [
    "\uE1D0",
    "\uE1D2",
    "\uE1D3",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
    "\uE1D5",
]


def _to_compound_note_codes(note_counts) -> List[str]:
    note_codes = []
    for i, note_count in enumerate(note_counts):
        for _ in range(note_count):
            note_codes.append(NOTE_CODES[i])
    return note_codes


def _to_compound_note_codes_alt(note_counts) -> List[str]:
    note_codes = []
    for i, note_count in enumerate(note_counts):
        for _ in range(note_count):
            note_codes.append(NOTE_CODES_ALT[i])
    return note_codes


def _to_note_counts(note_value) -> List[int]:
    """Return a list of counts of notes that sum to a note value.

    For example, a note value of 1.75 will be decoposed into a whole
    note, a half note and a quarter note, respectively. The return list
    is of length 12, where the first value corresponds to the count of
    double whole notes and the last value corresponds to the count of
    1024th notes.

    """
    if note_value < 2:
        base2 = bin(int(note_value * 1024))[2:]
        return [0] * (12 - len(base2)) + [int(bit) for bit in base2]
    first_bit, remainder = divmod(note_value, 2)
    base2 = bin(int(remainder * 1024))[2:]
    return (
        [int(first_bit)]
        + [0] * (11 - len(base2))
        + [int(bit) for bit in base2]
    )


def to_note_codes(note_value) -> List[str]:
    """Return a note value as a list of note codes.

    For example, a note value of 1.75 will be decoposed into a whole
    note, a half note and a quarter note, respectively.

    """
    note_code = COMMON_NOTE_CODES.get(note_value)
    if note_code is not None:
        return [note_code]
    note_counts = _to_note_counts(note_value)
    return _to_compound_note_codes(note_counts)


def to_note_codes_alt(note_value) -> List[str]:
    """Return a note value as a list of note codes with straight beams.

    This is useful for making chords as all the notes comes with a
    straight beam. For example, a note value of 1.75 will be decomposed
    into a whole note, a half note and a quarter note, respectively.

    """
    note_code = COMMON_NOTE_CODES_ALT.get(note_value)
    if note_code is not None:
        return [note_code]
    note_counts = _to_note_counts(note_value)
    return _to_compound_note_codes_alt(note_counts)


def get_time_signature_code(number: int) -> str:
    """Return a string for a time signature number."""
    if number > 10:
        div, mod = divmod(number, 10)
        return chr(57472 + div) + chr(57472 + mod)
    return chr(57472 + number)


def get_pitch_classes(fifths: int) -> List[int]:
    """Return a list of the root note of each pitch number."""
    if fifths >= 0:
        return [0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6]
    return [0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6]


def get_accidentals(fifths: int) -> List[Optional[int]]:
    """Return a list of the accidentals and the pitch classes."""
    if fifths >= 0:
        accidentals = [None, 1, None, 1, None, None, 1, None, 1, None, 1, None]
        if fifths > 0:
            sharps = [5, 0, 7, 2, 9, 4, 11]
            for idx in sharps[:fifths]:
                # No accidental for the sharp note
                accidentals[idx + 1] = None
                # Natural for the original note
                accidentals[idx] = 0
    elif fifths < 0:
        accidentals = [
            None,
            -1,
            None,
            -1,
            None,
            None,
            -1,
            None,
            -1,
            None,
            -1,
            None,
        ]
        flats = [11, 4, 9, 2, 7, 0, 5]
        for idx in flats[:(-fifths)]:
            # No accidental for the flat note
            accidentals[idx - 1] = None
            # Natural for the original note
            accidentals[idx] = 0

    return accidentals


class ScorePlotter:
    """A plotter that handles the score visualization.

    Attributes
    ----------
    fig : :class:`matplotlib.figure.Figure`
        Figure object to plot the score on.
    axes : :class:`matplotlib.axes.Axes`
        Axes object to plot the score on.
    resolution : int
        Time steps per quarter note.
    note_spacing : int, default: 4
        Spacing of notes.
    font_path : str or Path, optional
        Path to the music font. Defaults to the path to the downloaded
        Bravura font.
    font_scale : float, default: 140
        Font scaling factor for finetuning. The default value of 140 is
        optimized for the default Bravura font.

    """

    def __init__(
        self,
        fig: Figure,
        ax: Axes,
        resolution: int,
        note_spacing: int = None,
        font_path: Union[str, Path] = None,
        font_scale: float = None,
    ):
        self.fig = fig
        self.ax = ax
        self.resolution = resolution
        self.note_spacing = note_spacing if note_spacing is not None else 4
        if font_path is None:
            self.font_path = get_bravura_font_path()
        else:
            self.font_path = Path(font_path)
        self.font_scale = 140 if font_scale is None else font_scale

        # Check if font path exists
        if not self.font_path.exists():
            raise RuntimeError(
                "Music font not found. You could download it by "
                "`muspy.download_bravura_font()`."
            )

        # Set the axes to 1:1 aspect ratio
        self.ax.set_aspect("equal")

        # Turn off the axis
        self.ax.axis(False)

        # Initialize vertical boundaries
        self.left = 0.0
        self.right = 0.0
        self.bottom = 0.0
        self.top = 0.0

        # Initialize lists to store created objects
        self.staffs: List[Line2D] = []
        self.bar_lines: List[Line2D] = []
        self.final_bar_line = None
        self.clefs: List[Text] = []
        self.tempo_texts: List[Text] = []
        self.tempo_notes: List[Text] = []
        self.key_signatures: List[Text] = []
        self.time_signatures: List[Text] = []
        self.notes: List[Text] = []
        self.ties: List[Arc] = []

        # Initialize baseline position
        # (i.e., the y-coordinate of the first, lowest staff line)
        self._baseline = 0.0

        # Initialize horizontal position cursor
        self._pos = 0.0

        # Variables for handling notes
        self._force_new_note = True
        self._clef_offset = 0.0
        self._last_note_time = 0
        self._last_note_pos = 0.0
        self._last_note_y = 0.0
        self._splits_max = 0
        self._bottom_note_y = 0.0
        self._top_note_y = 0.0

        # Variables for handling key signatures
        self._pitch_classes = get_pitch_classes(0)
        self._accidentals = get_accidentals(0)

    def set_baseline(self, y):
        """Set baseline position (y-coordinate of first staff line)."""
        self._baseline = y
        self.top = max(self.top, self._baseline)
        self.bottom = max(self.bottom, self._baseline)
        self._bottom_note_y = self._baseline
        self._top_note_y = self._baseline + 4

    def adjust_fonts(self, scale: float = None):
        """Adjust the fonts."""
        if scale is None:
            scale = self.font_scale

        self.ax.set_xlim(self.left, self.right)
        self.ax.set_ylim(self.bottom, self.top)

        # Compute scaling factor
        ax_pos = self.ax.get_position()
        scaling_factor = (ax_pos.y1 - ax_pos.y0) / (self.top - self.bottom)

        # Set font size for normal texts
        fontsize = self.fig.get_figheight() * scaling_factor * scale
        for tempo_text in self.tempo_texts:
            tempo_text.set_fontsize(fontsize)
            tempo_text.set_fontfamily("serif")

        # Set music font for tempo notes
        prop_small = FontProperties(fname=self.font_path, size=fontsize)
        for tempo_note in self.tempo_notes:
            tempo_note.set_fontproperties(prop_small)

        # Set music font for music texts
        prop = FontProperties(fname=self.font_path, size=fontsize * 2)
        for key_signature in self.key_signatures:
            key_signature.set_fontproperties(prop)
        for time_signature in self.time_signatures:
            time_signature.set_fontproperties(prop)
        for clef in self.clefs:
            clef.set_fontproperties(prop)
        for note in self.notes:
            note.set_fontproperties(prop)

    def update_boundaries(
        self,
        left: float = None,
        right: float = None,
        bottom: float = None,
        top: float = None,
    ):
        """Update boundaries."""
        if left is not None:
            self.left = min(self.left, left)
        if right is not None:
            self.right = max(self.right, right)
        if bottom is not None:
            self.bottom = min(self.bottom, bottom)
        if top is not None:
            self.top = max(self.top, top)

    def plot_staffs(
        self, start: float = None, end: float = None
    ) -> List[Line2D]:
        """Plot the staffs."""
        if start is None:
            start = 0
        if end is None:
            end = self._pos

        staffs = []
        for y in range(5):
            staff = Line2D(
                [start, end],
                [self._baseline + y, self._baseline + y],
                color="k",
                linewidth=2,
            )
            self.ax.add_line(staff)
            staffs.append(staff)

        # Update boundaries
        self.update_boundaries(
            left=start - 1,
            right=end + 1,
            bottom=self._baseline - 1,
            top=self._baseline + 5,
        )
        return staffs

    def plot_bar_line(self) -> Line2D:
        """Plot a bar line."""
        bar_line = Line2D(
            (self._pos, self._pos),
            (self._baseline, self._baseline + 4),
            linewidth=2,
            color="k",
        )
        self.ax.add_line(bar_line)
        # Update boundaries
        self.update_boundaries(left=self._pos - 1, right=self._pos + 1)
        # Move position cursor
        self._pos += 1
        # Force the next note to be a new note
        self._force_new_note = True
        return bar_line

    def plot_final_bar_line(self) -> List[Artist]:
        """Plot an ending bar line."""
        bar_line = Line2D(
            (self._pos, self._pos),
            (self._baseline, self._baseline + 4),
            linewidth=2,
            color="k",
        )
        self.ax.add_line(bar_line)
        thick_bar_line = Rectangle(
            (self._pos + 0.5, self._baseline), 0.5, 4, color="k", linewidth=2,
        )
        self.ax.add_patch(thick_bar_line)
        # Update boundaries
        self.update_boundaries(left=self._pos - 1, right=self._pos + 1)
        # Move position cursor
        self._pos += 1
        return [bar_line, thick_bar_line]

    def plot_clef(self, kind="treble", octave=0) -> Text:
        """Plot a clef."""
        # Treble clef
        if kind.lower() == "treble":
            if octave == 0:
                clef = self.ax.text(self._pos, 1, "\uE050")
                self._clef_offset = 0
            elif octave == 1:
                clef = self.ax.text(self._pos, 1, "\uE053")
                self._clef_offset = -3.5
            elif octave == -1:
                clef = self.ax.text(self._pos, 1, "\uE052")
                self._clef_offset = 3.5
            elif octave == 2:
                clef = self.ax.text(self._pos, 1, "\uE054")
                self._clef_offset = -7
            elif octave == -2:
                clef = self.ax.text(self._pos, 1, "\uE051")
                self._clef_offset = 7
            else:
                raise ValueError(
                    "`octave` must be one of 0, ±1, ±2 for a treble clef."
                )
            # Update boundaries
            self.update_boundaries(
                bottom=self._baseline - 2, top=self._baseline + 6
            )

        # Alto clef
        elif kind.lower() == "alto":
            if octave == 0:
                clef = self.ax.text(self._pos, 2, "\uE05C")
                self._clef_offset = 3
            elif octave == -1:
                clef = self.ax.text(self._pos, 2, "\uE05D")
                self._clef_offset = 6.5
            else:
                raise ValueError(
                    "`octave` must be either 0 or -1 for an alto clef."
                )
            # Update boundaries
            self.update_boundaries(bottom=self._baseline - 1)

        # Bass clef
        elif kind.lower() == "bass":
            if octave == 0:
                clef = self.ax.text(self._pos, 3, "\uE062")
                self._clef_offset = 6
            elif octave == 1:
                clef = self.ax.text(self._pos, 3, "\uE065")
                self._clef_offset = 2.5
            elif octave == -1:
                clef = self.ax.text(self._pos, 3, "\uE064")
                self._clef_offset = 9.5
            elif octave == 2:
                clef = self.ax.text(self._pos, 3, "\uE066")
                self._clef_offset = -1
            elif octave == -2:
                clef = self.ax.text(self._pos, 3, "\uE063")
                self._clef_offset = 13
            else:
                raise ValueError(
                    "`octave` must be one of 0, ±1, ±2 for a bass clef."
                )
            # Update boundaries
            self.update_boundaries(bottom=self._baseline - 1)

        # Append clef to the correpsonding attribute
        self.clefs.append(clef)
        # Update boundaries
        self.update_boundaries(left=self._pos - 1, right=self._pos + 4)
        # Move position cursor
        self._pos += 4
        # Force the next note to be a new note
        self._force_new_note = True
        return clef

    def plot_tempo(self, qpm) -> List[Artist]:
        """Plot a tempo as a metronome mark."""
        note = self.ax.text(self._pos, 7, "\uE1D5", ha="right")
        self.tempo_notes.append(note)
        text = self.ax.text(self._pos, 7, " = " + str(int(qpm)))
        self.tempo_texts.append(text)
        # Update boundaries
        self.update_boundaries(
            left=self._pos - 2, right=self._pos + 4, top=self._baseline + 10
        )
        return [note, text]

    def plot_key_signature(self, root: int, mode: str):
        """Plot a key signature. Supports only major and minor keys."""
        # self._accidentals = _get_accidentals(0)
        # self._pitches = [0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6]
        # TODO: Support keys other than major and minor keys

    def plot_time_signature(
        self, numerator: int, denominator: int
    ) -> List[Text]:
        """Plot a time signature."""
        numerator_text = self.ax.text(
            self._pos + 1, 3, get_time_signature_code(numerator), ha="center"
        )
        self.time_signatures.append(numerator_text)
        denominator_text = self.ax.text(
            self._pos + 1, 1, get_time_signature_code(denominator), ha="center"
        )
        self.time_signatures.append(denominator_text)
        # Update boundaries
        self.update_boundaries(
            left=self._pos - 1,
            right=self._pos + 5,
            bottom=self._baseline - 2,
            top=self._baseline + 6,
        )
        # Move position cursor
        self._pos += 4
        # Force the next note to be a new note
        self._force_new_note = True
        return [numerator_text, denominator_text]

    def plot_note(
        self, time, duration, pitch
    ) -> Optional[Tuple[List[Text], List[Arc]]]:
        """Plot a note."""
        if time < self._last_note_time:
            warnings.warn(
                "Note times must be sorted. Skipped unsorted one.",
                RuntimeWarning,
            )
            return None

        # If force new note is set, reset all state variables
        if self._force_new_note:
            self._last_note_pos = self._pos
            self._splits_max = 0
            self._bottom_note_y = self._baseline
            self._top_note_y = self._baseline + 4

        # If not a chord, reset all state variables
        elif time > self._last_note_time:
            self._pos = (
                self._last_note_pos + self.note_spacing * self._splits_max
            )
            self._last_note_pos = self._pos
            self._splits_max = 0
            self._bottom_note_y = self._baseline
            self._top_note_y = self._baseline + 4

        # If not a chord, set position back to last note
        else:
            self._pos = self._last_note_pos

        # Compute note position
        octave, pitch_class = divmod(pitch, 12)
        note_y = (
            self._pitch_classes[pitch_class] / 2
            + (octave - 5) * 3.5
            - 1
            + self._clef_offset
        )

        # Get note codes
        if (
            not self._force_new_note
            and time == self._last_note_time
            and note_y + 3.5 > self._last_note_y
        ):
            note_codes = to_note_codes_alt(duration * 0.25 / self.resolution)
        else:
            note_codes = to_note_codes(duration * 0.25 / self.resolution)

        # Plot notes
        texts: List[Text] = []
        ties: List[Arc] = []
        for i, code in enumerate(note_codes):
            # Note without an accidental
            if self._accidentals[pitch_class] is None:
                text = self.ax.text(self._pos, note_y, code)
            # Note with a sharp
            elif self._accidentals[pitch_class] == 1:
                text = self.ax.text(self._pos - 1, note_y, "\uE262" + code)
            # Note with a flat
            elif self._accidentals[pitch_class] == -1:
                text = self.ax.text(self._pos - 1, note_y, "\uE260" + code)
            # Note with a natural
            elif self._accidentals[pitch_class] == 0:
                text = self.ax.text(self._pos - 1, note_y, "\uE261" + code)
            self.notes.append(text)
            texts.append(text)

            # Plot a slur if not the first note
            if i > 0:
                x_center = self._pos - self.note_spacing * 0.5 + 0.7
                # Upper half
                tie = Arc(
                    (x_center, note_y + 1.8),
                    1.4 * self.note_spacing,
                    1.4 * self.note_spacing,
                    theta1=240,
                    theta2=300,
                    linewidth=2,
                )
                self.ax.add_patch(tie)
                self.ties.append(tie)
            self._pos += self.note_spacing

        # Extend margins
        self._splits_max = max(self._splits_max, len(note_codes))
        self._last_note_time = time
        self._last_note_y = note_y

        # Plot ledger lines
        if note_y < self._bottom_note_y:
            for y in range(int(note_y), int(self._bottom_note_y)):
                self.ax.plot(
                    (
                        self._pos - self.note_spacing - 0.4,
                        self._pos - self.note_spacing + 1.7,
                    ),
                    (y, y),
                    linewidth=2,
                    color="k",
                )
            self._bottom_note_y = note_y

        elif note_y > self._top_note_y:
            for y in range(int(self._top_note_y) + 1, int(note_y) + 1):
                self.ax.plot(
                    (
                        self._pos - self.note_spacing - 0.4,
                        self._pos - self.note_spacing + 1.7,
                    ),
                    (y, y),
                    linewidth=2,
                    color="k",
                )
            self._top_note_y = note_y

        # Update boundaries
        self.update_boundaries(
            left=self._pos + self.note_spacing,
            right=self._last_note_pos - 1,
            bottom=self._baseline + note_y - 2.5,
            top=self._baseline + note_y + 5.5,
        )

        self._force_new_note = False

        return texts, ties

    def plot_object(self, obj):
        """Plot an object."""
        if isinstance(obj, Note):
            self.plot_note(obj.time, obj.duration, obj.pitch)
        elif isinstance(obj, Tempo):
            self.plot_tempo(obj.qpm)
        elif isinstance(obj, TimeSignature):
            self.plot_time_signature(obj.numerator, obj.denominator)
        elif isinstance(obj, KeySignature):
            self.plot_key_signature(obj.root, obj.mode)


def show_score(
    music: "Music",
    figsize: Tuple[float, float] = None,
    clef: str = "treble",
    clef_octave: int = 0,
    note_spacing: int = None,
    font_path: Union[str, Path] = None,
    font_scale: float = None,
) -> ScorePlotter:
    """Show score visualization.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to show.
    figsize : (float, float), optional
        Width and height in inches. Defaults to Matplotlib
        configuration.
    clef : {'treble', 'alto', 'bass'}, default: 'treble'
        Clef type.
    clef_octave : int, default: 0
        Clef octave.
    note_spacing : int, default: 4
        Spacing of notes.
    font_path : str or Path, optional
        Path to the music font. Defaults to the path to the downloaded
        Bravura font.
    font_scale : float, default: 140
        Font scaling factor for finetuning. The default value of 140 is
        optimized for the default Bravura font.

    Returns
    -------
    :class:`muspy.ScorePlotter`
        A ScorePlotter object that handles the score.

    """
    # Create a figure
    fig = plt.figure(figsize=figsize)

    # Add a full-size axes (will be resized at the end)
    ax = fig.add_axes((0, 0, 1, 1))

    # Create a score plotter
    plotter = ScorePlotter(
        fig,
        ax,
        resolution=music.resolution,
        note_spacing=note_spacing,
        font_path=font_path,
        font_scale=font_scale,
    )

    # Begining bar line
    plotter.plot_bar_line()

    # Clef
    plotter.plot_clef(kind=clef, octave=clef_octave)

    # Tempos, key signatures, time signatures and notes
    notes = []
    for track in music.tracks:
        notes += track.notes
    notes.sort(key=lambda x: (x.time, -x.pitch, x.duration))

    objects: List[Base] = []
    objects += music.key_signatures
    objects += music.time_signatures
    objects += music.tempos
    objects += notes
    objects.sort(key=attrgetter("time"))
    for obj in objects:
        plotter.plot_object(obj)

    # Final bar line
    plotter.plot_final_bar_line()

    # Staff lines
    plotter.plot_staffs()

    # Adjust fonts
    plotter.adjust_fonts(scale=font_scale)

    return plotter



"""Pitch-based representation output interface."""
from operator import attrgetter
from typing import TYPE_CHECKING, Union

import numpy as np
from numpy import ndarray

if TYPE_CHECKING:
    from ..music import Music


def to_pitch_representation(
    music: "Music",
    use_hold_state: bool = False,
    dtype: Union[np.dtype, type, str] = int,
) -> ndarray:
    """Encode a Music object into pitch-based representation.

    The pitch-based represetantion represents music as a sequence of
    pitch, rest and (optional) hold tokens. Only monophonic melodies are
    compatible with this representation. The output shape is T x 1,
    where T is the number of time steps. The values indicate whether the
    current time step is a pitch (0-127), a rest (128) or, optionally, a
    hold (129).

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to encode.
    use_hold_state : bool, default: False
        Whether to use a special state for holds.
    dtype : np.dtype, type or str, default: int
        Data type of the return array.

    Returns
    -------
    ndarray, shape=(?, 1)
        Encoded array in pitch-based representation.

    """
    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes are found
    if not notes:
        raise RuntimeError("No notes found.")

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    # Initialize the array
    length = max((note.end for note in notes))
    array = np.zeros((length, 1), dtype)

    # Fill the array with rests
    if use_hold_state:
        array.fill(128)

    # Encode note pitches
    for note in notes:
        if use_hold_state:
            array[note.time] = note.pitch
            array[note.time + 1 : note.time + note.duration] = 129
        else:
            array[note.time : note.time + note.duration] = note.pitch

    return array



"""Representation processors.

This module defines the processors for commonly used representations.

Classes
-------

- NoteRepresentationProcessor
- EventRepresentationProcessor
- PianoRollRepresentationProcessor
- PitchRepresentationProcessor

"""
from typing import Union

import numpy as np
from numpy import ndarray

from .inputs import (
    from_event_representation,
    from_note_representation,
    from_pianoroll_representation,
    from_pitch_representation,
)
from .music import Music
from .outputs import (
    to_event_representation,
    to_note_representation,
    to_pianoroll_representation,
    to_pitch_representation,
)

__all__ = [
    "NoteRepresentationProcessor",
    "EventRepresentationProcessor",
    "PianoRollRepresentationProcessor",
    "PitchRepresentationProcessor",
]


class NoteRepresentationProcessor:
    """Note-based representation processor.

    The note-based represetantion represents music as a sequence of
    (pitch, time, duration, velocity) tuples. For example, a note
    Note(time=0, duration=4, pitch=60, velocity=64) will be encoded as a
    tuple (0, 4, 60, 64). The output shape is L * D, where L is th
    number of notes and D is 4 when `encode_velocity` is True, otherwise
    D is 3. The values of the second dimension represent pitch, time,
    duration and velocity (discarded when `encode_velocity` is False).

    Attributes
    ----------
    use_start_end : bool, default: False
        Whether to use 'start' and 'end' to encode the timing rather
        than 'time' and 'duration'.
    encode_velocity : bool, default: True
        Whether to encode note velocities.
    dtype : dtype, type or str, default: int
        Data type of the return array.
    default_velocity : int, default: 64
        Default velocity value to use when decoding if `encode_velocity`
        is False.

    """

    def __init__(
        self,
        use_start_end: bool = False,
        encode_velocity: bool = True,
        dtype: Union[np.dtype, type, str] = int,
        default_velocity: int = 64,
    ):
        self.use_start_end = use_start_end
        self.encode_velocity = encode_velocity
        self.dtype = dtype
        self.default_velocity = default_velocity

    def encode(self, music: Music) -> ndarray:
        """Encode a Music object into note-based representation.

        Parameters
        ----------
        music : :class:`muspy.Music` object
            Music object to encode.

        Returns
        -------
        ndarray (np.uint8)
            Encoded array in note-based representation.

        See Also
        --------
        :func:`muspy.to_note_representation` :
            Convert a Music object into note-based representation.

        """
        return to_note_representation(
            music,
            use_start_end=self.use_start_end,
            encode_velocity=self.encode_velocity,
            dtype=self.dtype,
        )

    def decode(self, array: ndarray) -> Music:
        """Decode note-based representation into a Music object.

        Parameters
        ----------
        array : ndarray
            Array in note-based representation to decode. Cast to
            integer if not of integer type.

        Returns
        -------
        :class:`muspy.Music` object
            Decoded Music object.

        See Also
        --------
        :func:`muspy.from_note_representation` :
            Return a Music object converted from note-based
            representation.

        """
        return from_note_representation(
            array,
            use_start_end=self.use_start_end,
            encode_velocity=self.encode_velocity,
            default_velocity=self.default_velocity,
        )


class EventRepresentationProcessor:
    """Event-based representation processor.

    The event-based represetantion represents music as a sequence of
    events, including note-on, note-off, time-shift and velocity events.
    The output shape is M x 1, where M is the number of events. The
    values encode the events. The default configuration uses 0-127 to
    encode note-one events, 128-255 for note-off events, 256-355 for
    time-shift events, and 356 to 387 for velocity events.

    Attributes
    ----------
    use_single_note_off_event : bool, default: False
        Whether to use a single note-off event for all the pitches. If
        True, the note-off event will close all active notes, which can
        lead to lossy conversion for polyphonic music.
    use_end_of_sequence_event : bool, default: False
        Whether to append an end-of-sequence event to the encoded
        sequence.
    encode_velocity : bool, default: False
        Whether to encode velocities.
    force_velocity_event : bool, default: True
        Whether to add a velocity event before every note-on event. If
        False, velocity events are only used when the note velocity is
        changed (i.e., different from the previous one).
    max_time_shift : int, default: 100
        Maximum time shift (in ticks) to be encoded as an separate
        event. Time shifts larger than `max_time_shift` will be
        decomposed into two or more time-shift events.
    velocity_bins : int, default: 32
        Number of velocity bins to use.
    default_velocity : int, default: 64
        Default velocity value to use when decoding.

    """

    def __init__(
        self,
        use_single_note_off_event: bool = False,
        use_end_of_sequence_event: bool = False,
        encode_velocity: bool = False,
        force_velocity_event: bool = True,
        max_time_shift: int = 100,
        velocity_bins: int = 32,
        default_velocity: int = 64,
    ):
        self.use_single_note_off_event = use_single_note_off_event
        self.use_end_of_sequence_event = use_end_of_sequence_event
        self.encode_velocity = encode_velocity
        self.force_velocity_event = force_velocity_event
        self.max_time_shift = max_time_shift
        self.velocity_bins = velocity_bins
        self.default_velocity = default_velocity

    def encode(self, music: Music) -> ndarray:
        """Encode a Music object into event-based representation.

        Parameters
        ----------
        music : :class:`muspy.Music` object
            Music object to encode.

        Returns
        -------
        ndarray (np.uint16)
            Encoded array in event-based representation.

        See Also
        --------
        :func:`muspy.to_event_representation` :
            Convert a Music object into event-based representation.

        """
        return to_event_representation(
            music,
            use_single_note_off_event=self.use_single_note_off_event,
            use_end_of_sequence_event=self.use_end_of_sequence_event,
            encode_velocity=self.encode_velocity,
            force_velocity_event=self.force_velocity_event,
            max_time_shift=self.max_time_shift,
            velocity_bins=self.velocity_bins,
        )

    def decode(self, array: ndarray) -> Music:
        """Decode event-based representation into a Music object.

        Parameters
        ----------
        array : ndarray
            Array in event-based representation to decode. Cast to
            integer if not of integer type.

        Returns
        -------
        :class:`muspy.Music` object
            Decoded Music object.

        See Also
        --------
        :func:`muspy.from_event_representation` :
            Return a Music object converted from event-based
            representation.

        """
        return from_event_representation(
            array,
            use_single_note_off_event=self.use_single_note_off_event,
            use_end_of_sequence_event=self.use_end_of_sequence_event,
            max_time_shift=self.max_time_shift,
            velocity_bins=self.velocity_bins,
            default_velocity=self.default_velocity,
        )


class PitchRepresentationProcessor:
    """Pitch-based representation processor.

    The pitch-based represetantion represents music as a sequence of
    pitch, rest and (optional) hold tokens. Only monophonic melodies are
    compatible with this representation. The output shape is T x 1,
    where T is the number of time steps. The values indicate whether the
    current time step is a pitch (0-127), a rest (128) or, optionally, a
    hold (129).

    Attributes
    ----------
    use_hold_state : bool, default: False
        Whether to use a special state for holds.
    default_velocity : int, default: 64
        Default velocity value to use when decoding.

    """

    def __init__(
        self, use_hold_state: bool = False, default_velocity: int = 64,
    ):
        self.use_hold_state = use_hold_state
        self.default_velocity = default_velocity

    def encode(self, music: Music) -> ndarray:
        """Encode a Music object into pitch-based representation.

        Parameters
        ----------
        music : :class:`muspy.Music` object
            Music object to encode.

        Returns
        -------
        ndarray (np.uint8)
            Encoded array in pitch-based representation.

        See Also
        --------
        :func:`muspy.to_pitch_representation` :
            Convert a Music object into pitch-based representation.

        """
        return to_pitch_representation(
            music, use_hold_state=self.use_hold_state
        )

    def decode(self, array: ndarray) -> Music:
        """Decode pitch-based representation into a Music object.

        Parameters
        ----------
        array : ndarray
            Array in pitch-based representation to decode. Cast to
            integer if not of integer type.

        Returns
        -------
        :class:`muspy.Music` object
            Decoded Music object.

        See Also
        --------
        :func:`muspy.from_pitch_representation` :
            Return a Music object converted from pitch-based
            representation.

        """
        return from_pitch_representation(
            array,
            use_hold_state=self.use_hold_state,
            default_velocity=self.default_velocity,
        )


class PianoRollRepresentationProcessor:
    """Piano-roll representation processor.

    The piano-roll represetantion represents music as a time-pitch
    matrix, where the columns are the time steps and the rows are the
    pitches. The values indicate the presence of pitches at different
    time steps. The output shape is T x 128, where T is the number of
    time steps.

    Attributes
    ----------
    encode_velocity : bool, default: True
        Whether to encode velocities. If True, a binary-valued array
        will be return. Otherwise, an integer array will be return.
    default_velocity : int, default: 64
        Default velocity value to use when decoding if `encode_velocity`
        is False.

    """

    def __init__(
        self, encode_velocity: bool = True, default_velocity: int = 64,
    ):
        self.encode_velocity = encode_velocity
        self.default_velocity = default_velocity

    def encode(self, music: Music) -> ndarray:
        """Encode a Music object into piano-roll representation.

        Parameters
        ----------
        music : :class:`muspy.Music` object
            Music object to encode.

        Returns
        -------
        ndarray (np.uint8)
            Encoded array in piano-roll representation.

        See Also
        --------
        :func:`muspy.to_pianoroll_representation` :
            Convert a Music object into piano-roll representation.

        """
        return to_pianoroll_representation(
            music, encode_velocity=self.encode_velocity
        )

    def decode(self, array: ndarray) -> Music:
        """Decode piano-roll representation into a Music object.

        Parameters
        ----------
        array : ndarray
            Array in piano-roll representation to decode. Cast to
            integer if not of integer type. If `encode_velocity` is
            True, casted to boolean if not of boolean type.

        Returns
        -------
        :class:`muspy.Music` object
            Decoded Music object.

        See Also
        --------
        :func:`muspy.from_pianoroll_representation` :
            Return a Music object converted from piano-roll
            representation.

        """
        return from_pianoroll_representation(
            array,
            encode_velocity=self.encode_velocity,
            default_velocity=self.default_velocity,
        )



"""Audio output interface."""
import subprocess
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Union

import numpy as np
from numpy import ndarray

from ..external import get_musescore_soundfont_path
from .midi import write_midi

if TYPE_CHECKING:
    from ..music import Music


def _check_soundfont(soundfont_path):
    if soundfont_path is None:
        soundfont_path = get_musescore_soundfont_path()
    else:
        soundfont_path = Path(soundfont_path)
    if not soundfont_path.exists():
        raise RuntimeError(
            "Soundfont not found. Please download it by "
            "`muspy.download_musescore_soundfont()`."
        )
    return soundfont_path


def synthesize(
    music: "Music",
    soundfont_path: Union[str, Path] = None,
    rate: int = 44100,
    gain: float = None,
) -> ndarray:
    """Synthesize a Music object to raw audio.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to write.
    soundfont_path : str or Path, optional
        Path to the soundfount file. Defaults to the path to the
        downloaded MuseScore General soundfont.
    rate : int, default: 44100
        Sample rate (in samples per sec).
    gain : float, optional
        Master gain (`-g` option) for Fluidsynth. Defaults to 1/n,
        where n is the number of tracks. This can be used to prevent
        distortions caused by clipping.

    Returns
    -------
    ndarray, dtype=int16, shape=(?, 2)
        Synthesized waveform.

    """
    if gain is None:
        gain = 1 / len(music)

    # Check soundfont
    soundfont_path = _check_soundfont(soundfont_path)

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:

        # Write the Music object to a temporary MIDI file
        midi_path = Path(temp_dir) / "temp.mid"
        write_midi(midi_path, music)

        # Synthesize the MIDI file using fluidsynth
        result = subprocess.run(
            [
                "fluidsynth",
                "-T",
                "raw",
                "-F-",
                "-r",
                str(rate),
                "-g",
                str(gain),
                "-i",
                str(soundfont_path),
                str(midi_path),
            ],
            check=True,
            stdout=subprocess.PIPE,
        )

    # Decode bytes to waveform
    waveform = np.frombuffer(result.stdout, np.int16).reshape(-1, 2)

    return waveform


def write_audio(
    path: Union[str, Path],
    music: "Music",
    audio_format: str = None,
    soundfont_path: Union[str, Path] = None,
    rate: int = 44100,
    gain: float = None,
):
    """Write a Music object to an audio file.

    Supported formats include WAV, AIFF, FLAC and OGA.

    Parameters
    ----------
    path : str or Path
        Path to write the audio file.
    music : :class:`muspy.Music`
        Music object to write.
    audio_format : str, {'wav', 'aiff', 'flac', 'oga'}, optional
        File format to write. Defaults to infer from the extension.
    soundfont_path : str or Path, optional
        Path to the soundfount file. Defaults to the path to the
        downloaded MuseScore General soundfont.
    rate : int, default: 44100
        Sample rate (in samples per sec).
    gain : float, optional
        Master gain (`-g` option) for Fluidsynth. Defaults to 1/n,
        where n is the number of tracks. This can be used to prevent
        distortions caused by clipping.

    """
    if audio_format is None:
        audio_format = "auto"
    if gain is None:
        gain = 1 / len(music)

    # Check soundfont
    soundfont_path = _check_soundfont(soundfont_path)

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:

        # Write the Music object to a temporary MIDI file
        midi_path = Path(temp_dir) / "temp.mid"
        write_midi(midi_path, music)

        # Synthesize the MIDI file using fluidsynth
        subprocess.run(
            [
                "fluidsynth",
                "-ni",
                "-F",
                str(path),
                "-T",
                audio_format,
                "-r",
                str(rate),
                "-g",
                str(gain),
                str(soundfont_path),
                str(midi_path),
            ],
            check=True,
            stdout=subprocess.DEVNULL,
        )



"""Piano-roll visualization interface."""
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..music import Music


def show_pianoroll(music: "Music", **kwargs: Any):
    """Show pianoroll visualization."""
    multitrack = music.to_pypianoroll()
    return multitrack.plot(**kwargs)



"""Essen Folk Song Database."""
from .base import DatasetInfo, RemoteABCFolderDataset

# pylint: disable=line-too-long

_NAME = "Essen Folk Song Database"
_DESCRIPTION = """\
This is a collection of about 8000 European and Chinese folk songs written in \
Essen Associative Code (EsAC). This database was started under the direction \
of the late Helmut Scaffrath who invented the EsAC code. Seymour Shlien \
converted the database into abc music notation using a tcl/tk script that \
Seymour Shlien wrote. Damien Sagrillo and Ewa Dahlig-Turek made available \
several other collections (HAYDN.SM, IRL.SM, LUX.SM, LOT.SM, HAN1.SM and \
HAN2.SM which were not publicly available.)"""
_HOMEPAGE = "https://ifdo.ca/~seymour/runabc/esac/esacdatabase.html"


class EssenFolkSongDatabase(RemoteABCFolderDataset):
    """Essen Folk Song Database."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _sources = {
        "essen": {
            "filename": "esac.zip",
            "url": "https://ifdo.ca/~seymour/runabc/esac/esac.zip",
            "archive": True,
            "size": 1700981,
            "md5": "4636bd27b8ba4c57d2ef7db00d9eedc1",
            "sha256": "7957cf8f7a036dac9d807330548816967a13a4f598247e2a0f38aeccf04c7553",
        }
    }



"""Visualization interface."""
from typing import TYPE_CHECKING, Any

from .pianoroll import show_pianoroll
from .score import show_score

if TYPE_CHECKING:
    from ..music import Music


def show(music: "Music", kind: str, **kwargs: Any):
    """Show visualization.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to convert.
    kind : {'piano-roll', 'score'}
        Target representation.

    """
    if kind.lower() in ("piano-roll", "pianoroll", "piano roll"):
        return show_pianoroll(music, **kwargs)
    if kind.lower() == "score":
        return show_score(music, **kwargs)
    raise ValueError("`kind` must be one of 'piano-roll' and 'score'.")



"""EMOPIA Dataset."""
from pathlib import Path
from typing import Union

from ..classes import Annotation
from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "EMOPIA Dataset"
_DESCRIPTION = """\
EMOPIA (pronounced ‘yee-mò-pi-uh’) dataset is a shared multi-modal (audio and \
MIDI) database focusing on perceived emotion in pop piano music, to \
facilitate research on various tasks related to music emotion. The dataset \
contains 1,087 music clips from 387 songs and clip-level emotion labels \
annotated by four dedicated annotators."""
_HOMEPAGE = "https://annahung31.github.io/EMOPIA/"
_LICENSE = "Creative Commons Attribution 4.0 International License (CC-By 4.0)"
_CITATION = """\
@inproceedings{hung2021emopia,
  author={Hung, Hsiao-Tzu and Ching, Joann and Doh, Seungheon and Kim, Nabin and Nam, Juhan and Yang, Yi-Hsuan},
  title={{EMOPIA}: A Multi-Modal Pop Piano Dataset For Emotion Recognition and Emotion-based Music Generation},
  booktitle={Proceedings of the 22nd International Society for Music Information Retrieval Conference (ISMIR)},
  year=2021
}"""


class EMOPIADataset(RemoteFolderDataset):
    """EMOPIA Dataset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "emopia": {
            "filename": "EMOPIA_2.2.zip",
            "url": "https://zenodo.org/record/5257995/files/EMOPIA_2.2.zip",
            "archive": True,
            "md5": "bad5171786a4898f37fc2678e99afd94",
        }
    }

    _extension = "mid"

    def get_raw_filenames(self):
        """Return a list of raw filenames."""
        return sorted(
            (
                filename
                for filename in self.root.rglob("*." + self._extension)
                if not str(filename.relative_to(self.root)).startswith(
                    "_converted/"
                )
                and not str(filename.relative_to(self.root)).startswith(
                    "__MACOSX/"
                )
            )
        )

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        music = read_midi(self.root / filename)
        music.annotations = [parse_annotation(Path(filename).name)]
        return music


def parse_annotation(filename: str) -> Annotation:
    """Parse the annotation from the filename."""
    annotation = {
        "emo_class": filename[1],
        "YouTube_ID": filename[3:14],
        "seg_id": filename.split("_")[-1][:-4],
    }
    return Annotation(time=0, annotation=annotation)



"""MusicXML output interface."""
import os
from pathlib import Path
from typing import TYPE_CHECKING, Union

from music21.musicxml.archiveTools import compressXML

from .music21 import to_music21

if TYPE_CHECKING:
    from ..music import Music


def write_musicxml(
    path: Union[str, Path], music: "Music", compressed: bool = None
):
    """Write a Music object to a MusicXML file.

    Parameters
    ----------
    path : str or Path
        Path to write the MusicXML file.
    music : :class:`muspy.Music`
        Music object to write.
    compressed : bool, optional
        Whether to write to a compressed MusicXML file. If None, infer
        from the extension of the filename ('.xml' and '.musicxml' for
        an uncompressed file, '.mxl' for a compressed file).

    """
    score = to_music21(music)
    path = str(path)
    if compressed is None:
        if path.endswith((".xml", ".musicxml")):
            compressed = False
        elif path.endswith(".mxl"):
            compressed = True
        else:
            raise ValueError("Cannot infer file type from the extension.")

    if compressed:
        score.write("xml", path + ".temp.xml")
        compressXML(path + ".temp.xml", deleteOriginal=True)
        os.rename(path + ".temp.mxl", path)
    else:
        score.write("xml", path)



"""Input interfaces.

This module provides input interfaces for common symbolic music formats,
MusPy's native JSON and YAML formats, other symbolic music libraries and
commonly-used representations in music generation.

Functions
---------

- from_event_representation
- from_mido
- from_music21
- from_music21_opus
- from_note_representation
- from_object
- from_pianoroll_representation
- from_pitch_representation
- from_pretty_midi
- from_pypianoroll
- from_representation
- load
- load_json
- load_yaml
- read
- read_abc
- read_abc_string
- read_midi
- read_musicxml

Errors
------

- MIDIError
- MusicXMLError

"""
from .abc import read_abc, read_abc_string
from .event import from_event_representation
from .json import load_json
from .midi import MIDIError, from_mido, from_pretty_midi, read_midi
from .music21 import (
    from_music21,
    from_music21_opus,
    from_music21_part,
    from_music21_score,
)
from .musicxml import MusicXMLError, read_musicxml
from .note import from_note_representation
from .pianoroll import (
    from_pianoroll_representation,
    from_pypianoroll,
    from_pypianoroll_track,
)
from .pitch import from_pitch_representation
from .wrappers import from_object, from_representation, load, read
from .yaml import load_yaml

__all__ = [
    "MIDIError",
    "MusicXMLError",
    "from_event_representation",
    "from_mido",
    "from_music21",
    "from_music21_opus",
    "from_music21_part",
    "from_music21_score",
    "from_note_representation",
    "from_object",
    "from_pianoroll_representation",
    "from_pitch_representation",
    "from_pretty_midi",
    "from_pypianoroll",
    "from_pypianoroll_track",
    "from_representation",
    "load",
    "load_json",
    "load_yaml",
    "read",
    "read_abc",
    "read_abc_string",
    "read_midi",
    "read_musicxml",
]



"""Utility functions for dataset classes."""
import gzip
import hashlib
import lzma
import os
import os.path
import shutil
import tarfile
import zipfile
from pathlib import Path
from typing import Union
from urllib.request import urlretrieve

from tqdm import tqdm


class _ProgressBar:
    """A callable progress bar object.

    Code is adapted from https://stackoverflow.com/a/53643011.

    """

    def __init__(self):
        self.pbar = None

    def __call__(self, block_num, block_size, total_size):
        if self.pbar is None:
            self.pbar = tqdm(total=total_size)

        downloaded = block_num * block_size
        self.pbar.update(downloaded)


def check_size(path: Union[str, Path], size: int):
    """Check if the size of a file matches the expected one.

    Parameters
    ----------
    path : str or Path
        Path to the file.
    size : str
        Expected size of the file.

    """
    return Path(path).stat().st_size == size


def compute_md5(path: Union[str, Path], chunk_size: int):
    """Return the MD5 hash of a file, calculated chunk by chunk.

    Parameters
    ----------
    path : str or Path
        Path to the file.
    chunk_size : int
        Chunk size used to calculate the MD5 hash.

    """
    md5 = hashlib.md5()
    with open(str(path), "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            md5.update(chunk)
    return md5.hexdigest()


def check_md5(path: Union[str, Path], md5: str, chunk_size: int = 1024 * 1024):
    """Check if the MD5 hash of a file matches the expected one.

    Parameters
    ----------
    path : str or Path
        Path to the file.
    md5 : str
        Expected MD5 hash of the file.
    chunk_size : int, default: 2^20
        Chunk size used to compute the MD5 hash.

    """
    return compute_md5(path, chunk_size) == md5


def compute_sha256(path: Union[str, Path], chunk_size: int):
    """Return the MD5 checksum of a file, calculated chunk by chunk.

    Parameters
    ----------
    path : str or Path
        Path to the file.
    chunk_size : int
        Chunk size used to calculate the MD5 checksum.

    """
    sha256 = hashlib.sha256()
    with open(str(path), "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            sha256.update(chunk)
    return sha256.hexdigest()


def check_sha256(
    path: Union[str, Path], sha256: str, chunk_size: int = 1024 * 1024
):
    """Check if the sha256 hash of a file matches the expected one.

    Parameters
    ----------
    path : str or Path
        Path to the file.
    sha256 : str
        Expected sha256 hash of the file.
    chunk_size : int, default: 2^20
        Chunk size used to compute the sha256 hash.

    """
    return compute_sha256(path, chunk_size) == sha256


def download_url(
    url: str,
    path: Union[str, Path],
    overwrite: bool = False,
    size: int = None,
    md5: str = None,
    sha256: str = None,
    verbose: bool = True,
):
    """Download a file from a URL.

    Parameters
    ----------
    url : str
        URL to the file to download.
    path : str or Path
        Path to save the downloaded file.
    overwrite : bool, default: False
        Whether to overwrite existing downloaded file.
    size : int, optional
        Expected size of the downloaded file. Defaults to skip size
        check.
    md5 : str, optional
        Expected MD5 checksum of the downloaded file. Defaults to skip
        MD5 check.
    sha256 : str, optional
        Expected sha256 checksum of the downloaded file. Defaults to
        skip sha256 check.
    verbose : bool, default: True
        Whether to be verbose.

    """
    path = Path(path)
    if not overwrite and path.is_file():
        if size is not None and not check_size(path, size):
            raise RuntimeError(
                "Existing file has a different size from the expected one."
            )
        if md5 is not None and not check_md5(path, md5):
            raise RuntimeError(
                "Existing file has a different md5 hash from the expected one."
            )
        if sha256 is not None and not check_sha256(path, sha256):
            raise RuntimeError(
                "Existing file has a different sha256 hash from the expected "
                "one."
            )
        if verbose:
            print(f"Found existing downloaded file : {path} .")
        return

    # Download the file
    if verbose:
        print(f"Downloading source : {url} ...")
        urlretrieve(url, path, reporthook=_ProgressBar())
        print(f"Successfully downloaded source : {path} .")
    else:
        urlretrieve(url, path)

    # Run checks
    if size is not None and not check_size(path, size):
        raise RuntimeError(
            "Downloaded file has a different size from the expected one."
        )
    if md5 is not None and not check_md5(path, md5):
        raise RuntimeError(
            "Downloaded file has a different md5 hash from the expected one."
        )
    if sha256 is not None and not check_sha256(path, sha256):
        raise RuntimeError(
            "Downloaded file has a different sha256 hash from the expected "
            "one."
        )


def extract_archive(
    path: Union[str, Path],
    root: Union[str, Path] = None,
    kind: str = None,
    cleanup: bool = False,
    verbose: bool = True,
):
    """Extract an archive in ZIP, TAR, TGZ, TXZ, GZ or XZ format.

    Parameters
    ----------
    path : str or Path
        Path to the archive.
    root : str or Path, optional
        Root directory to save the extracted file. Defaults to the
        directory that contains the archive.
    kind : {'zip', 'tar', 'tgz', 'txz', 'gz', 'xz'}, optional
        Fromat of the archive. Defaults to infer from the extension.
    cleanup : bool, default: False
        Whether to remove the source archive after extraction.
    verbose : bool, default: True
        Whether to be verbose.

    """
    path = Path(path)
    root = Path(root) if root is not None else path.parent
    if kind is None:
        std_path = str(path)
        if std_path.endswith(".zip"):
            kind = "zip"
        elif std_path.endswith(".tar"):
            kind = "tar"
        elif std_path.endswith((".tar.gz", ".tgz")):
            kind = "tgz"
        elif std_path.endswith((".tar.xz", ".txz")):
            kind = "txz"
        elif std_path.endswith(".gz"):
            kind = "gz"
        elif std_path.endswith(".xz"):
            kind = "xz"
        else:
            raise ValueError(
                "Cannot infer file format from the extension (expect ZIP, "
                "TAR, TGZ, TXZ, GZ or XZ)."
            )

    # Extract archive
    if kind in ("zip", "tar", "tgz", "txz"):

        if verbose:
            print(f"Extracting archive : {path} ...")

        # zip file
        if kind == "zip":
            with zipfile.ZipFile(path, "r") as zip_file:
                zip_file.extractall(root)

        # tar file
        else:
            if kind == "tar":
                mode = "r"
            elif kind == "tgz":
                mode = "r:gz"
            elif kind == "txz":
                mode = "r:xz"
            with tarfile.open(path, mode) as f:
                f.extractall(root)

        if verbose:
            print(f"Successfully extracted archive : {root} .")

    elif kind in ("gz", "xz"):

        if verbose:
            print(f"Extracting archive : {path} ...")
        filename = root / path.stem

        # gzip file
        if kind == "gz":
            with gzip.open(path, "rb") as f_in, open(filename, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

        # xz file
        else:
            with lzma.open(path, "rb") as f_in_, open(filename, "wb") as f_out:
                shutil.copyfileobj(f_in_, f_out)

        if verbose:
            print(f"Successfully extracted archive : {filename} .")

    else:
        raise ValueError(
            "Expect `kind` to be one of 'zip', 'tar', 'tgz', 'txz', 'gz', "
            f"'xz' or , but got : {kind}."
        )

    # Cleanup source archive
    if cleanup:
        os.remove(path)
        if verbose:
            print("Removed source archive : {path} .")



"""Event-based representation output interface."""
from operator import attrgetter, itemgetter
from typing import TYPE_CHECKING, Iterable, List

import numpy as np
from bidict import bidict
from numpy import ndarray

if TYPE_CHECKING:
    from ..music import Music


def to_event_representation(
    music: "Music",
    use_single_note_off_event: bool = False,
    use_end_of_sequence_event: bool = False,
    encode_velocity: bool = False,
    force_velocity_event: bool = True,
    max_time_shift: int = 100,
    velocity_bins: int = 32,
    dtype=int,
) -> ndarray:
    """Encode a Music object into event-based representation.

    The event-based represetantion represents music as a sequence of
    events, including note-on, note-off, time-shift and velocity events.
    The output shape is M x 1, where M is the number of events. The
    values encode the events. The default configuration uses 0-127 to
    encode note-on events, 128-255 for note-off events, 256-355 for
    time-shift events, and 356 to 387 for velocity events.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to encode.
    use_single_note_off_event : bool, default: False
        Whether to use a single note-off event for all the pitches. If
        True, the note-off event will close all active notes, which can
        lead to lossy conversion for polyphonic music.
    use_end_of_sequence_event : bool, default: False
        Whether to append an end-of-sequence event to the encoded
        sequence.
    encode_velocity : bool, default: False
        Whether to encode velocities.
    force_velocity_event : bool, default: True
        Whether to add a velocity event before every note-on event. If
        False, velocity events are only used when the note velocity is
        changed (i.e., different from the previous one).
    max_time_shift : int, default: 100
        Maximum time shift (in ticks) to be encoded as an separate
        event. Time shifts larger than `max_time_shift` will be
        decomposed into two or more time-shift events.
    velocity_bins : int, default: 32
        Number of velocity bins to use.
    dtype : np.dtype, type or str, default: int
        Data type of the return array.

    Returns
    -------
    ndarray, shape=(?, 1)
        Encoded array in event-based representation.

    """
    if dtype is None:
        dtype = int

    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes is found
    if not notes and not use_end_of_sequence_event:
        raise RuntimeError("No notes found.")

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    # Compute offsets
    offset_note_on = 0
    offset_note_off = 128
    offset_time_shift = 129 if use_single_note_off_event else 256
    offset_velocity = offset_time_shift + max_time_shift
    if use_end_of_sequence_event:
        offset_eos = offset_velocity + velocity_bins

    # Collect note-related events
    note_events = []
    last_velocity = -1
    for note in notes:
        # Velocity event
        if encode_velocity:
            quantized_velocity = int(note.velocity * velocity_bins / 128)
            if force_velocity_event or quantized_velocity != last_velocity:
                note_events.append(
                    (note.time, offset_velocity + quantized_velocity)
                )
            last_velocity = quantized_velocity
        # Note on event
        note_events.append((note.time, offset_note_on + note.pitch))
        # Note off event
        if use_single_note_off_event:
            note_events.append((note.end, offset_note_off))
        else:
            note_events.append((note.end, offset_note_off + note.pitch))

    # Sort events by time
    note_events.sort(key=itemgetter(0))

    # Create a list for all events
    events = []
    # Initialize the time cursor
    time_cursor = 0
    # Iterate over note events
    for time, code in note_events:
        # If event time is after the time cursor, append tick shift
        # events
        if time > time_cursor:
            div, mod = divmod(time - time_cursor, max_time_shift)
            for _ in range(div):
                events.append(offset_time_shift + max_time_shift - 1)
            if mod > 0:
                events.append(offset_time_shift + mod - 1)
            events.append(code)
            time_cursor = time
        else:
            events.append(code)
    # Append the end-of-sequence event
    if use_end_of_sequence_event:
        events.append(offset_eos)

    return np.array(events, dtype=dtype).reshape(-1, 1)


class EventSequence(list):
    """A class for handling an event sequence.

    The EventSequence inherits from the builtin list. The elements are
    integer codes of the events defined by its `indexer` attribute. The
    corresponding events can be accessed by calling `events(idx)`.

    Attributes
    ----------
    indexer : bidict, optional
        Indexer that defines the mapping between events and their codes.

    """

    def __init__(self, iterable: Iterable = None, indexer: bidict = None):
        if iterable is not None:
            super().__init__(iterable)
        else:
            super().__init__()
        self.indexer = indexer if indexer is not None else bidict()

    def event(self, idx: int) -> str:
        """Return the event at a given index."""
        return self.indexer.inverse[self[idx]]

    def events(self) -> List[str]:
        """Return a list of all events."""
        return [self.indexer.inverse[elem] for elem in self]

    def append_event(self, event: str):
        """Append an event to the event sequence."""
        # pylint: disable=unsubscriptable-object
        self.append(self.indexer[event])

    def extend_events(self, events: List[str]):
        """Extend the event sequence by a list of events."""
        # pylint: disable=unsubscriptable-object
        self.extend(self.indexer[event] for event in events)

    def inverse(self, idx) -> str:
        """Return the corresponding event by its code."""
        return self.indexer.inverse[idx]


def get_default_indexer() -> bidict:
    """Return the default indexer."""
    indexer = {}
    idx = 0
    # Note-on events
    for i in range(128):
        indexer[f"note_on_{i}"] = idx
        idx += 1
    # Note-off events
    for i in range(128):
        indexer[f"note_off_{i}"] = idx
        idx += 1
    # Time-shift events
    for i in range(1, 101):
        indexer[f"time_shift_{i}"] = idx
        idx += 1
    return bidict(indexer)


class DefaultEventSequence(EventSequence):
    """A class for handling a MIDI-like event sequence.

    Attributes
    ----------
    indexer : bidict, optional
        Indexer that defines the mapping between events and their codes.

    """

    def __init__(self, iterable: Iterable = None, indexer: bidict = None):
        if indexer is not None:
            super().__init__(iterable, indexer)
        else:
            super().__init__(iterable, get_default_indexer())

    @classmethod
    def to_note_on_event(cls, pitch) -> str:
        """Return a note-on event for a given pitch."""
        return f"note_on_{pitch}"

    @classmethod
    def to_note_off_event(cls, pitch) -> str:
        """Return a note-off event for a given pitch."""
        return f"note_off_{pitch}"

    @classmethod
    def to_time_shift_events(cls, time_shift) -> List[str]:
        """Return a list of time-shift events for a given time-shift."""
        if time_shift <= 100:
            return [f"time_shift_{time_shift}"]
        events = []
        div, mod = divmod(time_shift, 100)
        for _ in range(div):
            events.append("time_shift_100")
        if mod > 0:
            events.append(f"time_shift_{mod}")
        return events


def to_default_event_sequence(music: "Music") -> DefaultEventSequence:
    """Return a Music object as a DefaultEventSequence object."""
    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes is found
    if not notes:
        raise RuntimeError("No notes found.")

    # Create a DefaultEventSequence object
    seq = DefaultEventSequence()

    # Collect events
    events = []
    for note in notes:
        events.append((note.time, seq.to_note_on_event(note.pitch)))
        events.append((note.end, seq.to_note_off_event(note.pitch)))

    # Sort the events by time
    events.sort(key=itemgetter(0))

    # Create event sequence
    last_event_time = 0
    for event in events:
        if event[0] > last_event_time:
            time_shift = event[0] - last_event_time
            seq.extend_events(seq.to_time_shift_events(time_shift))
        seq.append_event(event[1])
        last_event_time = event[0]

    return seq


def to_default_event_representation(music: "Music", dtype=int) -> ndarray:
    """Encode a Music object into the default event representation."""
    seq = to_default_event_sequence(music)
    return np.array(seq, dtype=dtype)


def get_performance_indexer() -> bidict:
    """Return the default indexer."""
    indexer = {}
    idx = 0
    # Note-on events
    for i in range(128):
        indexer[f"note_on_{i}"] = idx
        idx += 1
    # Note-off events
    for i in range(128):
        indexer[f"note_off_{i}"] = idx
        idx += 1
    # Time-shift events
    for i in range(1, 101):
        indexer[f"time_shift_{i}"] = idx
        idx += 1
    # Velocity events
    for i in range(32):
        indexer[f"velocity_{i}"] = idx
        idx += 1
    return bidict(indexer)


class PerformanceEventSequence(EventSequence):
    """A class for handling a MIDI-like event sequence.

    Attributes
    ----------
    indexer : bidict, optional
        Indexer that defines the mapping between events and their codes.

    """

    def __init__(self, iterable: Iterable = None, indexer: bidict = None):
        if indexer is not None:
            super().__init__(iterable, indexer)
        else:
            super().__init__(iterable, get_performance_indexer())

    @classmethod
    def to_note_on_event(cls, pitch) -> str:
        """Return a note-on event for a given pitch."""
        return f"note_on_{pitch}"

    @classmethod
    def to_note_off_event(cls, pitch) -> str:
        """Return a note-off event for a given pitch."""
        return f"note_off_{pitch}"

    @classmethod
    def to_velocity_event(cls, velocity) -> str:
        """Return a velocity event for a given velocity."""
        return f"velocity_{velocity//4}"

    @classmethod
    def to_time_shift_events(cls, time_shift) -> List[str]:
        """Return a list of time-shift events for a given time-shift."""
        if time_shift <= 100:
            return [f"time_shift_{time_shift}"]
        events = []
        div, mod = divmod(time_shift, 100)
        for _ in range(div):
            events.append("time_shift_100")
        if mod > 0:
            events.append(f"time_shift_{mod}")
        return events


def to_performance_event_sequence(music: "Music") -> PerformanceEventSequence:
    """Return a Music object as a PerformanceEventSequence object."""
    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes is found
    if not notes:
        raise RuntimeError("No notes found.")

    # Create a PerformanceEventSequence object
    seq = PerformanceEventSequence()

    # Collect events
    events = []
    for note in notes:
        events.append((note.time, seq.to_velocity_event(note.velocity)))
        events.append((note.time, seq.to_note_on_event(note.pitch)))
        events.append((note.end, seq.to_note_off_event(note.pitch)))

    # Sort the events by time
    events.sort(key=itemgetter(0))

    # Create event sequence
    last_event_time = 0
    for event in events:
        if event[0] > last_event_time:
            time_shift = event[0] - last_event_time
            seq.extend_events(seq.to_time_shift_events(time_shift))
        seq.append_event(event[1])
        last_event_time = event[0]

    return seq


def to_performance_event_representation(music: "Music", dtype=int) -> ndarray:
    """Encode a Music object into the performance event representation."""
    seq = to_performance_event_sequence(music)
    return np.array(seq, dtype=dtype)


def get_remi_indexer() -> bidict:
    """Return the REMI indexer."""
    indexer = {}
    idx = 0
    # Note-on events
    for i in range(128):
        indexer[f"note_on_{i}"] = idx
        idx += 1
    # Note-duration events
    for i in range(1, 65):
        indexer[f"note_duration_{i}"] = idx
        idx += 1
    # Position events
    for i in range(0, 24):
        indexer[f"position_{i}"] = idx
        idx += 1
    # Beat event
    indexer["beat"] = idx
    idx += 1
    # Tempo events
    for i in range(30, 210):
        indexer[f"tempo_{i}"] = idx
        idx += 1
    return bidict(indexer)


class REMIEventSequence(EventSequence):
    """A class for handling a MIDI-like event sequence.

    Attributes
    ----------
    indexer : bidict, optional
        Indexer that defines the mapping between events and their codes.

    Note
    ----
    Bar events are replaced by beat events. Chord events are currently
    not supported.

    """

    def __init__(self, iterable: Iterable = None, indexer: bidict = None):
        if indexer is not None:
            super().__init__(iterable, indexer)
        else:
            super().__init__(iterable, get_remi_indexer())

    @classmethod
    def to_note_on_event(cls, pitch) -> str:
        """Return a note-on event for a given pitch."""
        return f"note_on_{pitch}"

    @classmethod
    def to_note_duration_event(cls, duration) -> str:
        """Return a note-duration event for a given pitch."""
        return f"note_duration_{duration}"

    @classmethod
    def to_position_event(cls, position) -> str:
        """Return a position event for a given position."""
        return f"position_{position}"

    @classmethod
    def to_beat_event(cls) -> str:
        """Return a beat event."""
        return "beat"

    @classmethod
    def to_tempo_event(cls, tempo) -> str:
        """Return a position event for a given position."""
        return f"tempo_{int(tempo)}"


def to_remi_event_sequence(music: "Music") -> REMIEventSequence:
    """Return a Music object as a REMIEventSequence object."""
    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes is found
    if not notes:
        raise RuntimeError("No notes found.")

    # Create a REMIEventSequence object
    seq = REMIEventSequence()

    # Collect events
    events = []
    for tempo in music.tempos:
        events.append((tempo.time, seq.to_tempo_event(tempo.qpm)))
    for note in notes:
        events.append((note.time, seq.to_note_on_event(note.pitch)))
        events.append((note.end, seq.to_note_duration_event(note.duration)))

    # Sort the events by time
    events.sort(key=itemgetter(0))

    # Create event sequence
    last_beat = -1
    last_position = -1
    for event in events:
        beat, position = divmod(event[0], music.resolution)
        if beat > last_beat:
            seq.append_event(seq.to_beat_event())
        if position > last_position:
            seq.append_event(
                seq.to_position_event(24 * position // music.resolution)
            )
        seq.append_event(event[1])

    return seq


def to_remi_event_representation(music: "Music", dtype=int) -> ndarray:
    """Encode a Music object into the remi event representation."""
    seq = to_remi_event_sequence(music)
    return np.array(seq, dtype=dtype)


def get_indexer(preset=None) -> bidict:
    """Return a preset indexer."""
    if preset is None or preset.lower() == "midi":
        return get_default_indexer()
    if preset.lower() == "remi":
        return get_remi_indexer()
    if preset.lower() == "performance":
        return get_performance_indexer()
    raise ValueError(f"Unknown preset : {preset}")



"""ABC input interface."""
from pathlib import Path
from typing import List, Union

import music21.converter
from music21.stream import Opus

from ..music import DEFAULT_RESOLUTION, Music
from .music21 import from_music21_opus, from_music21_score


def read_abc_string(
    data_str: str, number: int = None, resolution=DEFAULT_RESOLUTION,
) -> Union[Music, List[Music]]:
    """Read ABC data into Music object(s) using music21 backend.

    Parameters
    ----------
    data_str : str
        ABC data to parse.
    number : int, optional
        Reference number of a specific tune to read (i.e., the 'X:'
        field). Defaults to read all tunes.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object(s).

    """
    # Parse the ABC data using music21
    parsed = music21.converter.parse(data_str, format="abc", number=number)

    # An ABC file can contain multiple songs
    if isinstance(parsed, Opus):
        # Convert the parsed music21 Opus object to MusPy Music objects
        music_list = from_music21_opus(parsed, resolution)

        # Set metadata
        for music in music_list:
            music.metadata.source_format = "abc"

        return music_list

    # Convert the parsed music21 Score object to a MusPy Music object
    music = from_music21_score(parsed, resolution)

    # Set metadata
    music.metadata.source_format = "abc"

    return music


def read_abc(
    path: Union[str, Path], number: int = None, resolution=DEFAULT_RESOLUTION,
) -> Union[Music, List[Music]]:
    """Return an ABC file into Music object(s) using music21 backend.

    Parameters
    ----------
    path : str or Path
        Path to the ABC file to read.
    number : int, optional
        Reference number of a specific tune to read (i.e., the 'X:'
        field). Defaults to read all tunes.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    list of :class:`muspy.Music`
        Converted Music object(s).

    """
    # Parse the ABC file using music21
    parsed = music21.converter.parse(path, format="abc", number=number)

    # An ABC file can contain multiple songs
    if isinstance(parsed, Opus):
        # Convert the parsed music21 Opus object to MusPy Music objects
        music_list = from_music21_opus(parsed, resolution)

        # Set metadata
        for music in music_list:
            music.metadata.source_filename = Path(path).name
            music.metadata.source_format = "abc"

        return music_list

    # Convert the parsed music21 Score object to a MusPy Music object
    music = from_music21_score(parsed, resolution)

    # Set metadata
    music.metadata.source_filename = Path(path).name
    music.metadata.source_format = "abc"

    return music



"""Music21 input interface."""
from operator import attrgetter
from typing import Dict, List, Tuple, Union

import numpy as np
from music21.instrument import partitionByInstrument
from music21.key import Key
from music21.key import KeySignature as M21KeySignature
from music21.stream import Opus, Part, Score, Stream

from ..classes import (
    Beat,
    Chord,
    KeySignature,
    Metadata,
    Note,
    Tempo,
    TimeSignature,
    Track,
)
from ..music import DEFAULT_RESOLUTION, Music


def parse_metadata(stream: Stream) -> Union[Metadata, None]:
    """Return metadata parsed from a music21 Stream object.

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to parse.

    Returns
    -------
    :class:`muspy.Metadata`
        Parsed metadata.

    """
    if stream.metadata is None:
        return None

    creators = []
    for contributor in stream.metadata.contributors:
        creators.extend(contributor.names)

    copyright_ = None
    for item in stream.metadata.all(skipContributors=True):
        if item[0] == "copyright":
            copyright_ = item[1]

    return Metadata(
        title=stream.metadata.title, creators=creators, copyright=copyright_,
    )


def parse_tempos(
    stream: Stream, resolution: int = DEFAULT_RESOLUTION
) -> List[Tempo]:
    """Return tempos parsed from a music21 Stream object.

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to parse.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    list of :class:`muspy.Tempo`
        Parsed tempos.

    """
    tempos = set()
    for start, _, metronome in stream.flat.metronomeMarkBoundaries():
        tempo = Tempo(
            time=round(float(start * resolution)),
            qpm=metronome.getQuarterBPM(),
        )
        tempos.add(tempo)
    return sorted(tempos, key=attrgetter("time"))


def parse_key_signatures(
    stream: Stream, resolution: int = DEFAULT_RESOLUTION
) -> List[KeySignature]:
    """Return key signatures parsed from a music21 Stream object.

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to parse.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    list of :class:`muspy.KeySignature`
        Parsed key signatures.

    """
    key_signatures = set()
    for item in stream.flat.getElementsByClass(M21KeySignature):
        if isinstance(item, Key):
            key_signatures.add(
                KeySignature(
                    time=round(float(item.offset * resolution)),
                    root=item.tonic.pitchClass,
                    mode=item.mode,
                    fifths=item.sharps,
                )
            )
        else:
            key_signatures.add(
                KeySignature(
                    time=round(float(item.offset * resolution)),
                    fifths=item.sharps,
                )
            )
    return sorted(key_signatures, key=attrgetter("time"))


def parse_time_signatures(
    stream: Stream, resolution: int = DEFAULT_RESOLUTION
) -> List[TimeSignature]:
    """Return time signatures parsed from a music21 Stream object.

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to parse.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    list of :class:`muspy.TimeSignature`
        Parsed time signatures.

    """
    time_signatures = set()
    for item in stream.flat.getTimeSignatures():
        time_signature = TimeSignature(
            time=round(float(item.offset * resolution)),
            numerator=item.numerator,
            denominator=item.denominator,
        )
        time_signatures.add(time_signature)
    return sorted(time_signatures, key=attrgetter("time"))


def parse_beats(
    stream: Stream,
    time_signatures: List[TimeSignature],
    resolution: int = DEFAULT_RESOLUTION,
) -> List[Beat]:
    """Return beats parsed from a music21 Stream object.

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to parse.
    time_signatures : list of :class:`muspy.TimeSignature`
        Time signature objects.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    list of :class:`muspy.Beat`
        Parsed beats.

    """
    beats: List[Beat] = []
    measure_offset_map = stream.measureOffsetMap()
    downbeats = list(
        round(float(offset * resolution))
        for offset in measure_offset_map.keys()
    )
    downbeats.sort()
    time_sign_idx = 0
    downbeat_idx = 0
    while downbeat_idx < len(downbeats):
        # Use next time signature
        if (
            time_sign_idx < len(time_signatures) - 1
            and downbeats[downbeat_idx] >= time_signatures[time_sign_idx].time
        ):
            time_sign_idx += 1
            continue
        # Set time signature
        time_sign = time_signatures[time_sign_idx]
        beat_resolution = resolution / (time_sign.denominator / 4)
        # Get the next downbeat
        if downbeat_idx < len(downbeats) - 1:
            end: float = downbeats[downbeat_idx + 1]
        else:
            end = (
                downbeats[downbeat_idx] + beat_resolution * time_sign.numerator
            )
        # Append beats
        for j, time in enumerate(
            np.arange(downbeats[downbeat_idx], end, beat_resolution)
        ):
            if j % time_sign.numerator == 0:
                beats.append(Beat(time=round(time), is_downbeat=True))
            else:
                beats.append(Beat(time=round(time), is_downbeat=False))
        downbeat_idx += 1
    return beats


def parse_notes_and_chords(
    stream: Stream, resolution: int = DEFAULT_RESOLUTION
) -> Tuple[List[Note], List[Chord]]:
    """Return notes and chords parsed from a music21 Stream object.

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to parse.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    list of :class:`muspy.Note`
        Parsed notes.
    list of :class:`muspy.Chord`
        Parsed chords.

    """
    notes: List[Note] = []
    chords: List[Chord] = []
    ties: Dict[int, int] = {}

    for item in stream.flat.notesAndRests:
        # Ignore rests
        if not item.isNote and not item.isChord:
            continue

        # Ignore grace notes
        if item.duration.isGrace:
            continue

        # Parse note
        time = round(float(item.offset * resolution))
        duration = round(float(item.quarterLength) * resolution)
        velocity = item.volume.velocity
        if velocity is not None:
            velocity = round(velocity)

        if item.isNote:
            pitch = int(item.pitch.midi)
            is_outgoing_tie = item.tie and (
                item.tie.type == "start" or item.tie.type == "continue"
            )
            if pitch in ties:
                note_idx = ties[pitch]
                notes[note_idx].duration += duration
                if is_outgoing_tie:
                    ties[pitch] = note_idx
                else:
                    del ties[pitch]
            else:
                note = Note(
                    time=time,
                    pitch=int(item.pitch.midi),
                    duration=duration,
                    velocity=velocity,
                )
                notes.append(note)
                if is_outgoing_tie:
                    ties[pitch] = len(notes) - 1

        elif item.isChord:
            chord = Chord(
                time=time,
                pitches=[int(note.pitch.midi) for note in item.notes],
                duration=duration,
                velocity=velocity,
            )
            chords.append(chord)

    return notes, chords


def parse_track(part: Part, resolution: int = DEFAULT_RESOLUTION) -> Track:
    """Return track parsed from a music21 Part object.

    Parameters
    ----------
    part : `music21.stream.Part`
        Part object to parse.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Track`
        Parsed track.

    """
    notes, chords = parse_notes_and_chords(part, resolution)

    instrument = part.getInstrument()

    if instrument.midiProgram is not None:
        program = instrument.midiProgram
    else:
        program = 0

    if instrument.midiChannel is not None:
        is_drum = instrument.midiChannel == 10
    else:
        is_drum = False

    return Track(
        program=program,
        is_drum=is_drum,
        name=part.partName,
        notes=notes,
        chords=chords,
    )


def from_music21_part(
    part: Part, resolution: int = DEFAULT_RESOLUTION
) -> Union[Track, List[Track]]:
    """Return a music21 Part object as Track object(s).

    Parameters
    ----------
    part : `music21.stream.Part`
        Part object to parse.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Track` or list of :class:`muspy.Track`
        Parsed track(s).

    """
    instruments = partitionByInstrument(part)
    if not instruments:
        return parse_track(part, resolution)
    return [parse_track(instrument, resolution) for instrument in instruments]


def from_music21_score(
    score: Score, resolution: int = DEFAULT_RESOLUTION
) -> Music:
    """Return a music21 Stream object as a Music object.

    Parameters
    ----------
    score : `music21.stream.Score`
        Score object to convert.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    tracks = []
    for part in score.parts:
        instruments = partitionByInstrument(part)
        if instruments:
            for instrument in instruments:
                tracks.append(parse_track(instrument))
        elif len(part.flat.notesAndRests) > 0:
            tracks.append(parse_track(part))

    time_signatures = parse_time_signatures(score, resolution)
    beats = parse_beats(score, time_signatures, resolution)

    return Music(
        metadata=parse_metadata(score),
        resolution=DEFAULT_RESOLUTION,
        tempos=parse_tempos(score),
        key_signatures=parse_key_signatures(score, resolution),
        time_signatures=time_signatures,
        beats=beats,
        tracks=tracks,
    )


def from_music21_opus(
    opus: Opus, resolution: int = DEFAULT_RESOLUTION
) -> List[Music]:
    """Return a music21 Opus object as a list of Music objects.

    Parameters
    ----------
    opus : `music21.stream.Opus`
        Opus object to convert.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    return [from_music21_score(score, resolution) for score in opus.scores]


def from_music21(
    stream: Stream, resolution: int = DEFAULT_RESOLUTION
) -> Union[Music, List[Music], Track, List[Track]]:
    """Return a music21 Stream object as Music or Track object(s).

    Parameters
    ----------
    stream : `music21.stream.Stream`
        Stream object to convert.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Music` or :class:`muspy.Track`
        Converted Music or Track object(s).

    """
    if isinstance(stream, Opus):
        return from_music21_opus(stream, resolution)
    elif isinstance(stream, Part):
        return from_music21_part(stream, resolution)
    else:
        return from_music21_score(stream, resolution)



"""Nottingham Database."""
from .base import DatasetInfo, RemoteABCFolderDataset

# pylint: disable=line-too-long

_NAME = "Nottingham Database"
_DESCRIPTION = """\
This is a collection of 1200 British and American folk tunes, (hornpipe, \
jigs, and etc.) that was created by Eric Foxley. The database was converted \
to ABC music notation format and was posted on http://abc.sourceforge.net/. \
The collection was edited by Seymour Shlien correcting missing beats during \
repeats and transitions."""
_HOMEPAGE = "https://ifdo.ca/~seymour/nottingham/nottingham.html"


class NottinghamDatabase(RemoteABCFolderDataset):
    """Nottingham Database."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _sources = {
        "nmd": {
            "filename": "nottingham_database.zip",
            "url": "http://ifdo.ca/~seymour/nottingham/nottingham_database.zip",
            "archive": True,
            "size": 142934,
            "md5": "f55c354aaf08bcb6e9b2b3b8d52e4df3",
            "sha256": "f79a4bffe78b16d630d4d69f9c62775a7aa246d0973c4d8714ab6c5139ff5a3b",
        }
    }



"""Wrapper function."""
from typing import Type

from .base import Dataset
from .emopia import EMOPIADataset
from .essen import EssenFolkSongDatabase
from .haydn import HaydnOp20Dataset
from .hymnal import HymnalDataset, HymnalTuneDataset
from .jsb import JSBChoralesDataset
from .lmd import (
    LakhMIDIAlignedDataset,
    LakhMIDIDataset,
    LakhMIDIMatchedDataset,
)
from .maestro import MAESTRODatasetV1, MAESTRODatasetV2, MAESTRODatasetV3
from .music21 import Music21Dataset
from .musicnet import MusicNetDataset
from .nes import NESMusicDatabase
from .nmd import NottinghamDatabase
from .wikifonia import WikifoniaDataset


def list_datasets():
    """Return all supported dataset classes as a list.

    Returns
    -------
    A list of all supported dataset classes.

    """
    return [
        EMOPIADataset,
        EssenFolkSongDatabase,
        HaydnOp20Dataset,
        HymnalDataset,
        HymnalTuneDataset,
        JSBChoralesDataset,
        LakhMIDIAlignedDataset,
        LakhMIDIDataset,
        LakhMIDIMatchedDataset,
        MAESTRODatasetV1,
        MAESTRODatasetV2,
        MAESTRODatasetV3,
        Music21Dataset,
        MusicNetDataset,
        NESMusicDatabase,
        NottinghamDatabase,
        WikifoniaDataset,
    ]


def get_dataset(key: str) -> Type[Dataset]:
    """Return a certain dataset class by key.

    Parameters
    ----------
    key : str
        Dataset key (case-insensitive).

    Returns
    -------
    The corresponding dataset class.

    """
    key = key.lower()
    if key == "emopia":
        return EMOPIADataset
    if key == "essen":
        return EssenFolkSongDatabase
    if key == "haydn":
        return HaydnOp20Dataset
    if key.startswith("hymnal"):
        if key == "hymnal":
            return HymnalDataset
        if key == "hymnal-tune":
            return HymnalTuneDataset
    if key == "jsb":
        return JSBChoralesDataset
    if key.startswith("lmd"):
        if key in ("lmd", "lmd-full"):
            return LakhMIDIDataset
        if key == "lmd-matched":
            return LakhMIDIMatchedDataset
        if key == "lmd-aligned":
            return LakhMIDIAlignedDataset
    if key.startswith("maestro"):
        if key in ("maestro", "maestro-v3"):
            return MAESTRODatasetV3
        if key == "maestro-v2":
            return MAESTRODatasetV2
        if key == "maestro-v1":
            return MAESTRODatasetV1
    if key == "music21":
        return Music21Dataset
    if key == "musicnet":
        return MusicNetDataset
    if key == "nes":
        return NESMusicDatabase
    if key == "nmd":
        return NottinghamDatabase
    if key == "wikifonia":
        return WikifoniaDataset
    raise ValueError(f"Unrecognized dataset key : {key}.")



"""Wrapper functions for output interface."""
from pathlib import Path
from typing import TYPE_CHECKING, TextIO, Union

from mido import MidiFile
from music21.stream import Stream
from numpy import ndarray
from pretty_midi import PrettyMIDI
from pypianoroll import Multitrack

from .audio import write_audio
from .event import to_event_representation
from .json import save_json
from .midi import to_mido, to_pretty_midi, write_midi
from .music21 import to_music21
from .musicxml import write_musicxml
from .note import to_note_representation
from .pianoroll import to_pianoroll_representation, to_pypianoroll
from .pitch import to_pitch_representation
from .yaml import save_yaml

if TYPE_CHECKING:
    from ..music import Music


def save(
    path: Union[str, Path, TextIO], music: "Music", kind: str = None, **kwargs,
):
    """Save a Music object loselessly to a JSON or a YAML file.

    This is a wrapper function for :func:`muspy.save_json` and
    :func:`muspy.save_yaml`.

    Parameters
    ----------
    path : str, Path or TextIO
        Path or file to save the data.
    music : :class:`muspy.Music`
        Music object to save.
    kind : {'json', 'yaml'}, optional
        Format to save. Defaults to infer from the extension.
    **kwargs
        Keyword arguments to pass to :func:`muspy.save_json` or
        :func:`muspy.save_yaml`.

    See Also
    --------
    :func:`muspy.save_json` : Save a Music object to a JSON file.
    :func:`muspy.save_yaml` : Save a Music object to a YAML file.
    :func:`muspy.write` :
        Write a Music object to a MIDI/MusicXML/ABC/audio file.

    Notes
    -----
    The conversion can be lossy if any nonserializable object is used
    (for example, an Annotation object, which can store data of any
    type).

    """
    if kind is None:
        if not isinstance(path, (str, Path)):
            raise ValueError("Cannot infer file format from a file object.")
        path_str = str(path).lower()
        if path_str.endswith((".json", ".json.gz")):
            kind = "json"
        elif path_str.endswith((".yaml", ".yml", ".yaml.gz", ".yml.gz")):
            kind = "yaml"
        else:
            raise ValueError(
                "Cannot infer file format from the extension (expect JSON or "
                "YAML)."
            )
    if kind.lower() == "json":
        return save_json(path, music, **kwargs)
    if kind.lower() == "yaml":
        return save_yaml(path, music, **kwargs)
    raise ValueError(
        f"Expect `kind` to be 'json' or 'yaml', but got : {kind}."
    )


def write(
    path: Union[str, Path], music: "Music", kind: str = None, **kwargs,
):
    """Write a Music object to a MIDI/MusicXML/ABC/audio file.

    Parameters
    ----------
    path : str or Path
        Path to write the file.
    music : :class:`muspy.Music`
        Music object to convert.
    kind : {'midi', 'musicxml', 'abc', 'audio'}, optional
        Format to save. Defaults to infer from the extension.

    See Also
    --------
    :func:`muspy.save` :
        Save a Music object loselessly to a JSON or a YAML file.

    """
    if kind is None:
        if str(path).lower().endswith((".mid", ".midi")):
            kind = "midi"
        elif (
            str(path).lower().endswith((".mxl", ".xml", ".mxml", ".musicxml"))
        ):
            kind = "musicxml"
        elif str(path).lower().endswith(".abc"):
            kind = "abc"
        elif str(path).lower().endswith(("wav", "aiff", "flac", "oga")):
            kind = "audio"
        else:
            raise ValueError(
                "Cannot infer file format from the extension (expect MIDI, "
                "MusicXML, ABC, WAV, AIFF, FLAC or OGA)."
            )
    if kind.lower() == "midi":
        return write_midi(path, music, **kwargs)
    if kind.lower() == "musicxml":
        return write_musicxml(path, music, **kwargs)
    if kind.lower() == "audio":
        return write_audio(path, music, **kwargs)
    raise ValueError(
        "Expect `kind` to be 'midi', 'musicxml', 'abc' or 'audio', but"
        f"got : {kind}."
    )


def to_object(
    music: "Music", kind: str, **kwargs
) -> Union[Stream, MidiFile, PrettyMIDI, Multitrack]:
    """Return a Music object as an object in other libraries.

    Supported classes are `music21.Stream`, :class:`mido.MidiTrack`,
    :class:`pretty_midi.PrettyMIDI` and :class:`pypianoroll.Multitrack`.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to convert.
    kind : str, {'music21', 'mido', 'pretty_midi', 'pypianoroll'}
        Target class.

    Returns
    -------
    `music21.Stream`, :class:`mido.MidiTrack`, \
            :class:`pretty_midi.PrettyMIDI` or \
            :class:`pypianoroll.Multitrack`
        Converted object.

    """
    if kind.lower() == "music21":
        return to_music21(music)
    if kind.lower() == "mido":
        return to_mido(music, **kwargs)
    if kind.lower() in ("pretty_midi", "prettymidi", "pretty-midi"):
        return to_pretty_midi(music)
    if kind.lower() == "pypianoroll":
        return to_pypianoroll(music)
    raise ValueError(
        "Expect `kind` to be 'music21', 'mido', 'pretty_midi' or "
        f"'pypianoroll', but got : {kind}."
    )


def to_representation(music: "Music", kind: str, **kwargs) -> ndarray:
    """Return a Music object in a specific representation.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to convert.
    kind : str, {'pitch', 'piano-roll', 'event', 'note'}
        Target representation.

    Returns
    -------
    array : ndarray
        Converted representation.

    """
    if kind.lower() in ("pitch", "pitch-based"):
        return to_pitch_representation(music, **kwargs)
    if kind.lower() in ("piano-roll", "pianoroll", "piano roll"):
        return to_pianoroll_representation(music, **kwargs)
    if kind.lower() in ("event", "event-based"):
        return to_event_representation(music, **kwargs)
    if kind.lower() in ("note", "note-based"):
        return to_note_representation(music, **kwargs)
    raise ValueError(
        "Expect `kind` to be 'pitch', 'pianoroll', 'event' or 'note', but"
        f"got : {kind}."
    )



"""Piano-roll input interface."""
from operator import attrgetter
from typing import List

import numpy as np
from numpy import ndarray
from pypianoroll import Multitrack
from pypianoroll import Track as PypianorollTrack

from ..classes import DEFAULT_VELOCITY, Metadata, Note, Tempo, Track
from ..music import DEFAULT_RESOLUTION, Music


def _pianoroll_to_notes(
    array: ndarray, encode_velocity: bool, default_velocity: int
) -> List[Note]:
    binarized = array > 0
    diff = np.diff(binarized, axis=0, prepend=0, append=0)
    notes = []
    for i in range(128):
        boundaries = np.nonzero(diff[:, i])[0]
        for note_idx in range(len(boundaries) // 2):
            start = boundaries[2 * note_idx]
            end = boundaries[2 * note_idx + 1]
            if encode_velocity:
                velocity = array[start, i]
            else:
                velocity = default_velocity
            note = Note(
                time=int(start),
                pitch=i,
                duration=int(end - start),
                velocity=int(velocity),
            )
            notes.append(note)

    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    return notes


def from_pypianoroll_track(
    track: PypianorollTrack, default_velocity: int = DEFAULT_VELOCITY
) -> Track:
    """Return a Pypianoroll Track object as a Track object.

    Parameters
    ----------
    track : :class:`pypianoroll.Track`
        Pypianoroll Track object to convert.
    default_velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Default velocity value to use when decoding.

    Returns
    -------
    :class:`muspy.Track`
        Converted track.

    """
    # Convert piano roll to notes
    notes = _pianoroll_to_notes(
        track.pianoroll, track.pianoroll.dtype == np.bool_, default_velocity
    )
    return Track(
        notes=notes,
        name=track.name if track.name else None,
        program=track.program,
        is_drum=track.is_drum,
    )


def from_pypianoroll(
    multitrack: Multitrack, default_velocity: int = DEFAULT_VELOCITY
) -> Music:
    """Return a Pypianoroll Multitrack object as a Music object.

    Parameters
    ----------
    multitrack : :class:`pypianoroll.Multitrack`
        Pypianoroll Multitrack object to convert.
    default_velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Default velocity value to use when decoding.

    Returns
    -------
    music : :class:`muspy.Music`
        Converted MusPy Music object.

    """
    # Tempos
    tempo_change_timings = np.diff(multitrack.tempo, prepend=-1).nonzero()[0]
    tempos = [
        Tempo(time, qpm=float(multitrack.tempo[time]))
        for time in tempo_change_timings
    ]
    # Tracks
    tracks = [
        from_pypianoroll_track(track, default_velocity)
        for track in multitrack.tracks
    ]
    return Music(
        resolution=int(multitrack.resolution),
        metadata=Metadata(title=multitrack.name) if multitrack.name else None,
        tempos=tempos,
        tracks=tracks,
    )


def from_pianoroll_representation(
    array: ndarray,
    resolution: int = DEFAULT_RESOLUTION,
    program: int = 0,
    is_drum: bool = False,
    encode_velocity: bool = True,
    default_velocity: int = DEFAULT_VELOCITY,
) -> Music:
    """Decode piano-roll representation into a Music object.

    Parameters
    ----------
    array : ndarray
        Array in piano-roll representation to decode.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.
    program : int, default: 0 (Acoustic Grand Piano)
        Program number, according to General MIDI specification [1].
        Valid values are 0 to 127.
    is_drum : bool, default: False
        Whether it is a percussion track.
    encode_velocity : bool, default: True
        Whether to encode velocities.
    default_velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Default velocity value to use when decoding. Only used when
        `encode_velocity` is True.

    Returns
    -------
    :class:`muspy.Music`
        Decoded Music object.

    References
    ----------
    [1] https://www.midi.org/specifications/item/gm-level-1-sound-set

    """
    if encode_velocity and not np.issubdtype(array.dtype, np.integer):
        raise TypeError(
            "Array must be of type int when `encode_velocity` is True."
        )
    if not encode_velocity and not np.issubdtype(array.dtype, np.bool):
        raise TypeError(
            "Array must be of type bool when `encode_velocity` is False."
        )

    # Convert piano roll to notes
    notes = _pianoroll_to_notes(array, encode_velocity, default_velocity)

    # Create the Track and Music objects
    track = Track(program=program, is_drum=is_drum, notes=notes)
    music = Music(resolution=resolution, tracks=[track])

    return music



"""Note-based representation input interface."""
from operator import attrgetter

import numpy as np
from numpy import ndarray

from ..classes import DEFAULT_VELOCITY, Note, Track
from ..music import DEFAULT_RESOLUTION, Music


def from_note_representation(
    array: ndarray,
    resolution: int = DEFAULT_RESOLUTION,
    program: int = 0,
    is_drum: bool = False,
    use_start_end: bool = False,
    encode_velocity: bool = True,
    default_velocity: int = DEFAULT_VELOCITY,
) -> Music:
    """Decode note-based representation into a Music object.

    Parameters
    ----------
    array : ndarray
        Array in note-based representation to decode.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.
    program : int, default: 0 (Acoustic Grand Piano)
        Program number, according to General MIDI specification [1].
        Valid values are 0 to 127.
    is_drum : bool, default: False
        Whether it is a percussion track.
    use_start_end : bool, default: False
        Whether to use 'start' and 'end' to encode the timing rather
        than 'time' and 'duration'.
    encode_velocity : bool, default: True
        Whether to encode note velocities.
    default_velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Default velocity value to use when decoding. Only used when
        `encode_velocity` is True.

    Returns
    -------
    :class:`muspy.Music`
        Decoded Music object.

    References
    ----------
    [1] https://www.midi.org/specifications/item/gm-level-1-sound-set

    """
    if not np.issubdtype(array.dtype, np.integer):
        raise TypeError("Array must be of type int.")

    notes = []
    velocity = default_velocity
    for note_tuple in array:
        if encode_velocity:
            velocity = note_tuple[3]

        if use_start_end:
            duration = note_tuple[2] - note_tuple[0]
        else:
            duration = note_tuple[2]

        notes.append(
            Note(
                time=int(note_tuple[0]),
                pitch=int(note_tuple[1]),
                duration=int(duration),
                velocity=int(velocity),
            )
        )

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    # Create the Track and Music objects
    track = Track(program=program, is_drum=is_drum, notes=notes)
    music = Music(resolution=resolution, tracks=[track])

    return music



"""Music class.

This module defines the core class of MusPy---the Music class, a
universal container for symbolic music.

Classes
-------

- Music

Variables
---------

- DEFAULT_RESOLUTION

"""
from collections import OrderedDict
from math import ceil, floor
from pathlib import Path
from typing import Any, Callable, List, Union

import numpy as np
from mido import MidiFile
from music21.stream import Stream
from numpy import ndarray
from pretty_midi import PrettyMIDI
from pypianoroll import Multitrack

from .base import ComplexBase
from .classes import (
    Annotation,
    Beat,
    KeySignature,
    Lyric,
    Metadata,
    Tempo,
    TimeSignature,
    Track,
)
from .outputs import save, synthesize, to_object, to_representation, write
from .visualization import show

__all__ = ["Music", "DEFAULT_RESOLUTION"]

DEFAULT_RESOLUTION = 24

# pylint: disable=super-init-not-called


class Music(ComplexBase):
    """A universal container for symbolic music.

    This is the core class of MusPy. A Music object can be constructed
    in the following ways.

    - :meth:`muspy.Music`: Construct by setting values for attributes.
    - :meth:`muspy.Music.from_dict`: Construct from a dictionary that
      stores the attributes and their values as key-value pairs.
    - :func:`muspy.read`: Read from a MIDI, a MusicXML or an ABC file.
    - :func:`muspy.load`: Load from a JSON or a YAML file saved by
      :func:`muspy.save`.
    - :func:`muspy.from_object`: Convert from a `music21.Stream`, a
      :class:`mido.MidiFile`, a :class:`pretty_midi.PrettyMIDI` or a
      :class:`pypianoroll.Multitrack` object.

    Attributes
    ----------
    metadata : :class:`muspy.Metadata`, default: `Metadata()`
        Metadata.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.
    tempos : list of :class:`muspy.Tempo`, default: []
        Tempo changes.
    key_signatures : list of :class:`muspy.KeySignature`, default: []
        Key signatures changes.
    time_signatures : list of :class:`muspy.TimeSignature`, default: []
        Time signature changes.
    beats : list of :class:`muspy.Beat`, default: []
        Beats.
    lyrics : list of :class:`muspy.Lyric`, default: []
        Lyrics.
    annotations : list of :class:`muspy.Annotation`, default: []
        Annotations.
    tracks : list of :class:`muspy.Track`, default: []
        Music tracks.

    Note
    ----
    Indexing a Music object returns the track of a certain index. That
    is, ``music[idx]`` returns ``music.tracks[idx]``. Length of a Music
    object is the number of tracks. That is, ``len(music)``  returns
    ``len(music.tracks)``.

    """

    _attributes = OrderedDict(
        [
            ("metadata", Metadata),
            ("resolution", int),
            ("tempos", Tempo),
            ("key_signatures", KeySignature),
            ("time_signatures", TimeSignature),
            ("beats", Beat),
            ("lyrics", Lyric),
            ("annotations", Annotation),
            ("tracks", Track),
        ]
    )
    _optional_attributes = [
        "metadata",
        "resolution",
        "tempos",
        "key_signatures",
        "time_signatures",
        "beats",
        "lyrics",
        "annotations",
        "tracks",
    ]
    _list_attributes = [
        "tempos",
        "key_signatures",
        "time_signatures",
        "beats",
        "lyrics",
        "annotations",
        "tracks",
    ]

    def __init__(
        self,
        metadata: Metadata = None,
        resolution: int = None,
        tempos: List[Tempo] = None,
        key_signatures: List[KeySignature] = None,
        time_signatures: List[TimeSignature] = None,
        beats: List[Beat] = None,
        lyrics: List[Lyric] = None,
        annotations: List[Annotation] = None,
        tracks: List[Track] = None,
    ):
        self.metadata = metadata if metadata is not None else Metadata()
        self.resolution = (
            resolution if resolution is not None else DEFAULT_RESOLUTION
        )
        self.tempos = tempos if tempos is not None else []
        self.key_signatures = (
            key_signatures if key_signatures is not None else []
        )
        self.time_signatures = (
            time_signatures if time_signatures is not None else []
        )
        self.beats = beats if beats is not None else []
        self.lyrics = lyrics if lyrics is not None else []
        self.annotations = annotations if annotations is not None else []
        self.tracks = tracks if tracks is not None else []

    def __len__(self) -> int:
        return len(self.tracks)

    def __getitem__(self, key: int) -> Track:
        return self.tracks[key]

    def __setitem__(self, key: int, value: Track):
        self.tracks[key] = value

    def get_end_time(self, is_sorted: bool = False) -> int:
        """Return the the time of the last event in all tracks.

        This includes tempos, key signatures, time signatures, note
        offsets, lyrics and annotations.

        Parameters
        ----------
        is_sorted : bool, default: False
            Whether all the list attributes are sorted.

        """

        def _get_end_time(list_):
            if not list_:
                return 0
            if is_sorted:
                return list_[-1].time
            return max(item.time for item in list_)

        if self.tracks:
            track_end_time = max(
                track.get_end_time(is_sorted) for track in self.tracks
            )
        else:
            track_end_time = 0

        end_time = max(
            _get_end_time(self.tempos),
            _get_end_time(self.key_signatures),
            _get_end_time(self.time_signatures),
            _get_end_time(self.beats),
            _get_end_time(self.lyrics),
            _get_end_time(self.annotations),
            track_end_time,
        )

        return end_time

    def get_real_end_time(self, is_sorted: bool = False) -> float:
        """Return the end time in realtime.

        This includes tempos, key signatures, time signatures, note
        offsets, lyrics and annotations. Assume 120 qpm (quarter notes
        per minute) if no tempo information is available.

        Parameters
        ----------
        is_sorted : bool, default: False
            Whether all the list attributes are sorted.

        """
        # Get symbolic end time
        end_time = self.get_end_time(is_sorted=is_sorted)

        # If no tempo information is available, assume 120 qpm
        if not self.tempos:
            return 0.5 * end_time / self.resolution

        # Compute the real end time
        position = 0.0
        qpm = 120.0
        factor = 60.0 / self.resolution
        real_end_time = 0.0
        for tempo in self.tempos:
            real_end_time += (tempo.time - position) * factor / qpm
            position = tempo.time
            qpm = tempo.qpm
        real_end_time += (end_time - position) * factor / qpm

        return real_end_time

    def infer_beats(self) -> List[Beat]:
        """Infer beats from the time signature changes.

        This assumes that there is a downbeat at each time signature
        change (this is not always true, e.g., for a pickup measure).

        Returns
        -------
        list of :class:`muspy.Beat`
            List of beats inferred from the time signature changes.
            Return an empty list if no time signature is found.

        """
        beats: List[Beat] = []
        for i, time_sign in enumerate(self.time_signatures):
            if i == len(self.time_signatures) - 1:
                end = self.get_end_time()
            else:
                end = self.time_signatures[i + 1].time
            beat_resolution = self.resolution / (time_sign.denominator / 4)
            for j, time in enumerate(
                np.arange(time_sign.time, end, beat_resolution)
            ):
                if j % time_sign.numerator == 0:
                    beats.append(Beat(time=round(time), is_downbeat=True))
                else:
                    beats.append(Beat(time=round(time), is_downbeat=False))
        return beats

    def adjust_resolution(
        self,
        target: int = None,
        factor: float = None,
        rounding: Union[str, Callable] = "round",
    ) -> "Music":
        """Adjust resolution and timing of all time-stamped objects.

        Parameters
        ----------
        target : int, optional
            Target resolution.
        factor : int or float, optional
            Factor used to adjust the resolution based on the formula:
            `new_resolution = old_resolution * factor`. For example, a
            factor of 2 double the resolution, and a factor of 0.5 halve
            the resolution.
        rounding : {'round', 'ceil', 'floor'} or callable, default:
        'round'
            Rounding mode.

        Returns
        -------
        Object itself.

        """
        if self.resolution is None:
            raise TypeError("`resolution` must be given.")
        if self.resolution < 0:
            raise ValueError("`resolution` must be positive.")

        if target is None and factor is None:
            raise ValueError("One of `target` and `factor` must be given.")
        if target is not None and factor is not None:
            raise ValueError("Only one of `target` and `factor` can be given.")

        if rounding is None or rounding == "round":
            rounding = round
        elif rounding == "ceil":
            rounding = ceil
        elif rounding == "floor":
            rounding = floor
        elif isinstance(rounding, str):
            raise ValueError(f"Unrecognized rounding mode : {rounding} .")

        if target is not None:
            if not isinstance(target, int):
                raise TypeError("`target` must be an integer.")
            target_ = int(target)
            factor_ = target / self.resolution

        if factor is not None:
            new_resolution = float(self.resolution * factor)
            if not new_resolution.is_integer():
                raise ValueError(
                    f"`factor` must be a factor of the original resolution "
                    f"{self.resolution}, but got : {factor}."
                )
            factor_ = float(factor)
            target_ = int(new_resolution)

        self.resolution = int(target_)
        self.adjust_time(lambda time: rounding(time * factor_))  # type: ignore
        return self

    def clip(self, lower: int = 0, upper: int = 127) -> "Music":
        """Clip the velocity of each note for each track.

        Parameters
        ----------
        lower : int, default: 0
            Lower bound.
        upper : int, default: 127
            Upper bound.

        Returns
        -------
        Object itself.

        """
        for track in self.tracks:
            track.clip(lower, upper)
        return self

    def transpose(self, semitone: int) -> "Music":
        """Transpose all the notes by a number of semitones.

        Parameters
        ----------
        semitone : int
            Number of semitones to transpose the notes. A positive value
            raises the pitches, while a negative value lowers the
            pitches.

        Returns
        -------
        Object itself.

        Notes
        -----
        Drum tracks are skipped.

        """
        for track in self.tracks:
            if not track.is_drum:
                track.transpose(semitone)
        return self

    def save(self, path: Union[str, Path], kind: str = None, **kwargs: Any):
        """Save loselessly to a JSON or a YAML file.

        Refer to :func:`muspy.save` for full documentation.

        """
        return save(path, self, kind=kind, **kwargs)

    def save_json(self, path: Union[str, Path], **kwargs: Any):
        """Save loselessly to a JSON file.

        Refer to :func:`muspy.save_json` for full documentation.

        """
        return save(path, self, kind="json", **kwargs)

    def save_yaml(self, path: Union[str, Path]):
        """Save loselessly to a YAML file.

        Refer to :func:`muspy.save_yaml` for full documentation.

        """
        return save(path, self, kind="yaml")

    def write(self, path: Union[str, Path], kind: str = None, **kwargs: Any):
        """Write to a MIDI, a MusicXML, an ABC or an audio file.

        Refer to :func:`muspy.write` for full documentation.

        """
        return write(path, self, kind=kind, **kwargs)

    def write_midi(self, path: Union[str, Path], **kwargs: Any):
        """Write to a MIDI file.

        Refer to :func:`muspy.write_midi` for full documentation.

        """
        return write(path, self, kind="midi", **kwargs)

    def write_musicxml(self, path: Union[str, Path], **kwargs: Any):
        """Write to a MusicXML file.

        Refer to :func:`muspy.write_musicxml` for full documentation.

        """
        return write(path, self, kind="musicxml", **kwargs)

    def write_abc(self, path: Union[str, Path], **kwargs: Any):
        """Write to an ABC file.

        Refer to :func:`muspy.write_abc` for full documentation.

        """
        return write(path, self, kind="abc", **kwargs)

    def write_audio(self, path: Union[str, Path], **kwargs: Any):
        """Write to an audio file.

        Refer to :func:`muspy.write_audio` for full documentation.

        """
        return write(path, self, kind="audio", **kwargs)

    def to_object(self, kind: str, **kwargs: Any):
        """Return as an object in other libraries.

        Refer to :func:`muspy.to_object` for full documentation.

        """
        return to_object(self, kind=kind, **kwargs)

    def to_music21(self, **kwargs: Any) -> Stream:
        """Return as a Stream object.

        Refer to :func:`muspy.to_music21` for full documentation.

        """
        return to_object(self, kind="music21", **kwargs)

    def to_mido(self, **kwargs: Any) -> MidiFile:
        """Return as a MidiFile object.

        Refer to :func:`muspy.to_mido` for full documentation.

        """
        return to_object(self, kind="mido", **kwargs)

    def to_pretty_midi(self, **kwargs: Any) -> PrettyMIDI:
        """Return as a PrettyMIDI object.

        Refer to :func:`muspy.to_pretty_midi` for full documentation.

        """
        return to_object(self, kind="pretty_midi", **kwargs)

    def to_pypianoroll(self, **kwargs: Any) -> Multitrack:
        """Return as a Multitrack object.

        Refer to :func:`muspy.to_pypianoroll` for full documentation.

        """
        return to_object(self, kind="pypianoroll", **kwargs)

    def to_representation(self, kind: str, **kwargs: Any) -> ndarray:
        """Return in a specific representation.

        Refer to :func:`muspy.to_representation` for full documentation.

        """
        return to_representation(self, kind=kind, **kwargs)

    def to_pitch_representation(self, **kwargs: Any) -> ndarray:
        """Return in pitch-based representation.

        Refer to :func:`muspy.to_pitch_representation` for full
        documentation.

        """
        return to_representation(self, kind="pitch", **kwargs)

    def to_pianoroll_representation(self, **kwargs: Any) -> ndarray:
        """Return in piano-roll representation.

        Refer to :func:`muspy.to_pianoroll_representation` for full
        documentation.

        """
        return to_representation(self, kind="piano-roll", **kwargs)

    def to_event_representation(self, **kwargs: Any) -> ndarray:
        """Return in event-based representation.

        Refer to :func:`muspy.to_event_representation` for full
        documentation.

        """
        return to_representation(self, kind="event", **kwargs)

    def to_note_representation(self, **kwargs: Any) -> ndarray:
        """Return in note-based representation.

        Refer to :func:`muspy.to_note_representation` for full
        documentation.

        """
        return to_representation(self, kind="note", **kwargs)

    def show(self, kind: str, **kwargs: Any):
        """Show visualization.

        Refer to :func:`muspy.show` for full documentation.

        """
        return show(self, kind, **kwargs)

    def show_score(self, **kwargs: Any):
        """Show score visualization.

        Refer to :func:`muspy.show_score` for full documentation.

        """
        return show(self, kind="score", **kwargs)

    def show_pianoroll(self, **kwargs: Any):
        """Show pianoroll visualization.

        Refer to :func:`muspy.show_pianoroll` for full documentation.

        """
        return show(self, kind="piano-roll", **kwargs)

    def synthesize(self, **kwargs) -> ndarray:
        """Synthesize a Music object to raw audio.

        Refer to :func:`muspy.synthesize` for full documentation.

        """
        return synthesize(self, **kwargs)



"""Evaluation metrics."""
import math

import numpy as np
from numpy import ndarray

from ..music import Music


def n_pitches_used(music: Music) -> int:
    """Return the number of unique pitches used.

    Drum tracks are ignored.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    int
        Number of unique pitch used.

    See Also
    --------
    :func:`muspy.n_pitch_class_used` :
        Compute the number of unique pitch classes used.

    """
    count = 0
    is_used = [False] * 128
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            if not is_used[note.pitch]:
                is_used[note.pitch] = True
                count += 1
    return count


def n_pitch_classes_used(music: Music) -> int:
    """Return the number of unique pitch classes used.

    Drum tracks are ignored.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    int
        Number of unique pitch classes used.

    See Also
    --------
    :func:`muspy.n_pitches_used` :
        Compute the number of unique pitches used.

    """
    count = 0
    is_used = [False] * 12
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            pitch_class = note.pitch % 12
            if not is_used[pitch_class]:
                is_used[pitch_class] = True
                count += 1
    return count


def pitch_range(music: Music) -> int:
    """Return the pitch range.

    Drum tracks are ignored. Return zero if no note is found.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    int
        Pitch range.

    """
    if not music.tracks:
        return 0
    if not any(len(track.notes) > 0 for track in music.tracks):
        return 0

    highest = 0
    lowest = 127
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            if note.pitch > highest:
                highest = note.pitch
            if note.pitch < lowest:
                lowest = note.pitch
    return highest - lowest


def empty_beat_rate(music: Music) -> float:
    r"""Return the ratio of empty beats.

    The empty-beat rate is defined as the ratio of the number of empty
    beats (where no note is played) to the total number of beats. Return
    NaN if song length is zero. This metric is also implemented in
    Pypianoroll [1].

    .. math:: empty\_beat\_rate = \frac{\#(empty\_beats)}{\#(beats)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    float
        Empty-beat rate.

    See Also
    --------
    :func:`muspy.empty_measure_rate` :
        Compute the ratio of empty measures.

    References
    ----------
    1. Hao-Wen Dong, Wen-Yi Hsiao, and Yi-Hsuan Yang, “Pypianoroll: Open
       Source Python Package for Handling Multitrack Pianorolls,” in
       Late-Breaking Demos of the 18th International Society for Music
       Information Retrieval Conference (ISMIR), 2018.

    """
    length = max(track.get_end_time() for track in music.tracks)
    if length < 1:
        return math.nan

    n_beats = length // music.resolution + 1
    is_empty = [True] * n_beats
    count = 0
    for track in music.tracks:
        for note in track.notes:
            start = note.time // music.resolution
            end = note.end // music.resolution
            for beat in range(start, end + 1):
                if is_empty[beat]:
                    is_empty[beat] = False
                    count += 1
    return 1 - (count / n_beats)


def empty_measure_rate(music: Music, measure_resolution: int) -> float:
    r"""Return the ratio of empty measures.

    The empty-measure rate is defined as the ratio of the number of
    empty measures (where no note is played) to the total number of
    measures. Note that this metric only works for songs with a constant
    time signature. Return NaN if song length is zero. This metric is
    used in [1].

    .. math::
        empty\_measure\_rate = \frac{\#(empty\_measures)}{\#(measures)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.
    measure_resolution : int
        Time steps per measure.

    Returns
    -------
    float
        Empty-measure rate.

    See Also
    --------
    :func:`muspy.empty_beat_rate` : Compute the ratio of empty beats.

    References
    ----------
    1. Hao-Wen Dong, Wen-Yi Hsiao, Li-Chia Yang, and Yi-Hsuan Yang,
       "MuseGAN: Multi-track sequential generative adversarial networks
       for symbolic music generation and accompaniment," in Proceedings
       of the 32nd AAAI Conference on Artificial Intelligence (AAAI),
       2018.

    """
    length = max(track.get_end_time() for track in music.tracks)
    if length < 1:
        return math.nan

    n_measures = length // measure_resolution + 1
    is_empty = [True] * n_measures
    count = 0
    for track in music.tracks:
        for note in track.notes:
            start = note.time // measure_resolution
            end = note.end // measure_resolution
            for measure in range(start, end + 1):
                if is_empty[measure]:
                    is_empty[measure] = False
                    count += 1
    return 1 - (count / n_measures)


def _get_pianoroll(music: Music) -> ndarray:
    """Return the binary pianoroll matrix."""
    length = max(track.get_end_time() for track in music.tracks)
    pianoroll = np.zeros((length, 128), bool)
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            pianoroll[note.time : note.end, note.pitch] = 1
    return pianoroll


def polyphony(music: Music) -> float:
    r"""Return the average number of pitches being played concurrently.

    The polyphony is defined as the average number of pitches being
    played at the same time, evaluated only at time steps where at least
    one pitch is on. Drum tracks are ignored. Return NaN if no note is
    found.

    .. math::
        polyphony = \frac{
            \#(pitches\_when\_at\_least\_one\_pitch\_is\_on)
        }{
            \#(time\_steps\_where\_at\_least\_one\_pitch\_is\_on)
        }

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    float
        Polyphony.

    See Also
    --------
    :func:`muspy.polyphony_rate` :
        Compute the ratio of time steps where multiple pitches are on.

    """
    pianoroll = _get_pianoroll(music)
    denominator = np.count_nonzero(pianoroll.sum(1) > 0)
    if denominator < 1:
        return math.nan
    return pianoroll.sum() / denominator


def polyphony_rate(music: Music, threshold: int = 2) -> float:
    r"""Return the ratio of time steps where multiple pitches are on.

    The polyphony rate is defined as the ratio of the number of time
    steps where multiple pitches are on to the total number of time
    steps. Drum tracks are ignored. Return NaN if song length is zero.
    This metric is used in [1], where it is called `polyphonicity`.

    .. math::
        polyphony\_rate = \frac{
            \#(time\_steps\_where\_multiple\_pitches\_are\_on)
        }{
            \#(time\_steps)
        }

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.
    threshold : int, default: 2
        Threshold of number of pitches to count into the numerator.

    Returns
    -------
    float
        Polyphony rate.

    See Also
    --------
    :func:`muspy.polyphony` :
        Compute the average number of pitches being played at the same
        time.

    References
    ----------
    1. Hao-Wen Dong, Wen-Yi Hsiao, Li-Chia Yang, and Yi-Hsuan Yang,
       "MuseGAN: Multi-track sequential generative adversarial networks
       for symbolic music generation and accompaniment," in Proceedings
       of the 32nd AAAI Conference on Artificial Intelligence (AAAI),
       2018.

    """
    pianoroll = _get_pianoroll(music)
    if len(pianoroll) < 1:
        return math.nan
    return np.count_nonzero(pianoroll.sum(1) > threshold) / len(pianoroll)


def _get_scale(root: int, mode: str) -> ndarray:
    """Return the scale mask for a specific root."""
    if mode == "major":
        c_scale = np.array([1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], bool)
    elif mode == "minor":
        c_scale = np.array([1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1, 0], bool)
    else:
        raise ValueError("`mode` must be either 'major' or 'minor'.")
    return np.roll(c_scale, root)


def pitch_in_scale_rate(music: Music, root: int, mode: str) -> float:
    r"""Return the ratio of pitches in a certain musical scale.

    The pitch-in-scale rate is defined as the ratio of the number of
    notes in a certain scale to the total number of notes. Drum tracks
    are ignored. Return NaN if no note is found. This metric is used in
    [1].

    .. math::
        pitch\_in\_scale\_rate = \frac{\#(notes\_in\_scale)}{\#(notes)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.
    root : int
        Root of the scale.
    mode : str, {'major', 'minor'}
        Mode of the scale.

    Returns
    -------
    float
        Pitch-in-scale rate.

    See Also
    --------
    :func:`muspy.scale_consistency` :
        Compute the largest pitch-in-class rate.

    References
    ----------
    1. Hao-Wen Dong, Wen-Yi Hsiao, Li-Chia Yang, and Yi-Hsuan Yang,
       "MuseGAN: Multi-track sequential generative adversarial networks
       for symbolic music generation and accompaniment," in Proceedings
       of the 32nd AAAI Conference on Artificial Intelligence (AAAI),
       2018.

    """
    scale = _get_scale(root, mode.lower())
    note_count = 0
    in_scale_count = 0
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            note_count += 1
            if scale[note.pitch % 12]:
                in_scale_count += 1
    if note_count < 1:
        return math.nan
    return in_scale_count / note_count


def scale_consistency(music: Music) -> float:
    r"""Return the largest pitch-in-scale rate.

    The scale consistency is defined as the largest pitch-in-scale rate
    over all major and minor scales. Drum tracks are ignored. Return NaN
    if no note is found. This metric is used in [1].

    .. math::
        scale\_consistency = \max_{root, mode}{
            pitch\_in\_scale\_rate(root, mode)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    float
        Scale consistency.

    See Also
    --------
    :func:`muspy.pitch_in_scale_rate` :
        Compute the ratio of pitches in a certain musical scale.

    References
    ----------
    1. Olof Mogren, "C-RNN-GAN: Continuous recurrent neural networks
       with adversarial training," in NeuIPS Workshop on Constructive
       Machine Learning, 2016.

    """
    max_in_scale_rate = 0.0
    for mode in ("major", "minor"):
        for root in range(12):
            rate = pitch_in_scale_rate(music, root, mode)
            if math.isnan(rate):
                return math.nan
            if rate > max_in_scale_rate:
                max_in_scale_rate = rate
    return max_in_scale_rate


def _get_drum_pattern(res: int, meter: str) -> ndarray:
    """Return the drum pattern mask of a specific meter."""
    drum_pattern = np.zeros(res, dtype=bool)
    drum_pattern[0] = 1
    if meter == "duple":
        if res % 4 == 0:
            drum_pattern[:: (res // 4)] = 1
        if res % 2 == 0:
            drum_pattern[:: (res // 2)] = 1
    elif meter == "triple":
        if res % 3 == 0:
            drum_pattern[:: (res // 3)] = 1
    else:
        raise ValueError("Only duple and triple meters are supported.")
    return drum_pattern


def drum_in_pattern_rate(music: Music, meter: str) -> float:
    r"""Return the ratio of drum notes in a certain drum pattern.

    The drum-in-pattern rate is defined as the ratio of the number of
    notes in a certain scale to the total number of notes. Only drum
    tracks are considered. Return NaN if no drum note is found. This
    metric is used in [1].

    .. math::
        drum\_in\_pattern\_rate = \frac{
            \#(drum\_notes\_in\_pattern)}{\#(drum\_notes)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.
    meter : str, {'duple', 'triple'}
        Meter of the drum pattern.

    Returns
    -------
    float
        Drum-in-pattern rate.

    See Also
    --------
    :func:`muspy.drum_pattern_consistency` :
        Compute the largest drum-in-pattern rate.

    References
    ----------
    1. Hao-Wen Dong, Wen-Yi Hsiao, Li-Chia Yang, and Yi-Hsuan Yang,
       "MuseGAN: Multi-track sequential generative adversarial networks
       for symbolic music generation and accompaniment," in Proceedings
       of the 32nd AAAI Conference on Artificial Intelligence (AAAI),
       2018.

    """
    drum_pattern = _get_drum_pattern(music.resolution, meter.lower())
    note_count = 0
    in_pattern_count = 0
    for track in music.tracks:
        if not track.is_drum:
            continue
        for note in track.notes:
            note_count += 1
            if drum_pattern[note.time % music.resolution]:
                in_pattern_count += 1
    if note_count < 1:
        return math.nan
    return in_pattern_count / note_count


def drum_pattern_consistency(music: Music) -> float:
    r"""Return the largest drum-in-pattern rate.

    The drum pattern consistency is defined as the largest
    drum-in-pattern rate over duple and triple meters. Only drum tracks
    are considered. Return NaN if no drum note is found.

    .. math::
        drum\_pattern\_consistency = \max_{meter}{
            drum\_in\_pattern\_rate(meter)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    float
        Drum pattern consistency.

    See Also
    --------
    :func:`muspy.drum_in_pattern_rate` :
        Compute the ratio of drum notes in a certain drum pattern.

    """
    drum_in_duple_pattern_rate = drum_in_pattern_rate(music, "duple")
    if math.isnan(drum_in_duple_pattern_rate):
        return math.nan
    drum_in_triple_pattern_rate = drum_in_pattern_rate(music, "triple")

    if drum_in_duple_pattern_rate > drum_in_triple_pattern_rate:
        return drum_in_duple_pattern_rate
    return drum_in_triple_pattern_rate


def _entropy(prob):
    with np.errstate(divide="ignore", invalid="ignore"):
        return -np.nansum(prob * np.log2(prob))


def pitch_entropy(music: Music) -> float:
    r"""Return the entropy of the normalized note pitch histogram.

    The pitch entropy is defined as the Shannon entropy of the
    normalized note pitch histogram. Drum tracks are ignored. Return NaN
    if no note is found.

    .. math::
        pitch\_entropy = -\sum_{i = 0}^{127}{
            P(pitch=i) \log_2 P(pitch=i)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    float
        Pitch entropy.

    See Also
    --------
    :func:`muspy.pitch_class_entropy` :
        Compute the entropy of the normalized pitch class histogram.

    """
    counter = np.zeros(128)
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            counter[note.pitch] += 1
    denominator = counter.sum()
    if denominator < 1:
        return math.nan
    prob = counter / denominator
    return _entropy(prob)


def pitch_class_entropy(music: Music) -> float:
    r"""Return the entropy of the normalized note pitch class histogram.

    The pitch class entropy is defined as the Shannon entropy of the
    normalized note pitch class histogram. Drum tracks are ignored.
    Return NaN if no note is found. This metric is used in [1].

    .. math::
        pitch\_class\_entropy = -\sum_{i = 0}^{11}{
            P(pitch\_class=i) \times \log_2 P(pitch\_class=i)}

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.

    Returns
    -------
    float
        Pitch class entropy.

    See Also
    --------
    :func:`muspy.pitch_entropy` :
        Compute the entropy of the normalized pitch histogram.

    References
    ----------
    1. Shih-Lun Wu and Yi-Hsuan Yang, "The Jazz Transformer on the Front
       Line: Exploring the Shortcomings of AI-composed Music through
       Quantitative Measures”, in Proceedings of the 21st International
       Society for Music Information Retrieval Conference, 2020.

    """
    counter = np.zeros(12)
    for track in music.tracks:
        if track.is_drum:
            continue
        for note in track.notes:
            counter[note.pitch % 12] += 1
    denominator = counter.sum()
    if denominator < 1:
        return math.nan
    prob = counter / denominator
    return _entropy(prob)


def groove_consistency(music: Music, measure_resolution: int) -> float:
    r"""Return the groove consistency.

    The groove consistency is defined as the mean hamming distance of
    the neighboring measures.

    .. math::
        groove\_consistency = 1 - \frac{1}{T - 1} \sum_{i = 1}^{T - 1}{
            d(G_i, G_{i + 1})}

    Here, :math:`T` is the number of measures, :math:`G_i` is the binary
    onset vector of the :math:`i`-th measure (a one at position that has
    an onset, otherwise a zero), and :math:`d(G, G')` is the hamming
    distance between two vectors :math:`G` and :math:`G'`. Note that
    this metric only works for songs with a constant time signature.
    Return NaN if the number of measures is less than two. This metric
    is used in [1].

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to evaluate.
    measure_resolution : int
        Time steps per measure.

    Returns
    -------
    float
        Groove consistency.

    References
    ----------
    1. Shih-Lun Wu and Yi-Hsuan Yang, "The Jazz Transformer on the Front
       Line: Exploring the Shortcomings of AI-composed Music through
       Quantitative Measures”, in Proceedings of the 21st International
       Society for Music Information Retrieval Conference, 2020.

    """
    length = max(track.get_end_time() for track in music.tracks)
    if measure_resolution < 1:
        raise ValueError("Measure resolution must be a positive integer.")

    n_measures = (length // measure_resolution) + 1
    if n_measures < 2:
        return math.nan

    groove_patterns = np.zeros((n_measures, measure_resolution), bool)

    for track in music.tracks:
        for note in track.notes:
            measure, position = divmod(note.time, measure_resolution)
            if not groove_patterns[measure, position]:
                groove_patterns[measure, position] = 1

    hamming_distance = np.count_nonzero(
        groove_patterns[:-1] != groove_patterns[1:]
    )

    return 1 - hamming_distance / (measure_resolution * (n_measures - 1))



"""Dataset classes.

This module provides an easy-to-use dataset management system. Each
supported dataset in MusPy comes with a class inherited from the base
MusPy Dataset class. It also provides interfaces to PyTorch and
TensorFlow for creating input pipelines for machine learning.

Base Classes
------------

- ABCFolderDataset
- Dataset
- DatasetInfo
- FolderDataset
- RemoteABCFolderDataset
- RemoteDataset
- RemoteFolderDataset
- RemoteMusicDataset
- MusicDataset

Dataset Classes
---------------
- EssenFolkSongDatabase
- EMOPIADataset
- HaydnOp20Dataset
- HymnalDataset
- HymnalTuneDataset
- JSBChoralesDataset
- LakhMIDIAlignedDataset
- LakhMIDIDataset
- LakhMIDIMatchedDataset
- MAESTRODatasetV1
- MAESTRODatasetV2
- Music21Dataset
- MusicNetDataset
- NESMusicDatabase
- NottinghamDatabase
- WikifoniaDataset

"""

from .base import (
    ABCFolderDataset,
    Dataset,
    DatasetInfo,
    FolderDataset,
    MusicDataset,
    RemoteABCFolderDataset,
    RemoteDataset,
    RemoteFolderDataset,
    RemoteMusicDataset,
)
from .emopia import EMOPIADataset
from .essen import EssenFolkSongDatabase
from .haydn import HaydnOp20Dataset
from .hymnal import HymnalDataset, HymnalTuneDataset
from .jsb import JSBChoralesDataset
from .lmd import (
    LakhMIDIAlignedDataset,
    LakhMIDIDataset,
    LakhMIDIMatchedDataset,
)
from .maestro import MAESTRODatasetV1, MAESTRODatasetV2, MAESTRODatasetV3
from .music21 import Music21Dataset
from .musicnet import MusicNetDataset
from .nes import NESMusicDatabase
from .nmd import NottinghamDatabase
from .wikifonia import WikifoniaDataset
from .wrapper import get_dataset, list_datasets

__all__ = [
    "ABCFolderDataset",
    "Dataset",
    "DatasetInfo",
    "EMOPIADataset",
    "EssenFolkSongDatabase",
    "FolderDataset",
    "HaydnOp20Dataset",
    "HymnalDataset",
    "HymnalTuneDataset",
    "JSBChoralesDataset",
    "LakhMIDIAlignedDataset",
    "LakhMIDIDataset",
    "LakhMIDIMatchedDataset",
    "MAESTRODatasetV1",
    "MAESTRODatasetV2",
    "MAESTRODatasetV3",
    "Music21Dataset",
    "MusicDataset",
    "MusicNetDataset",
    "NESMusicDatabase",
    "NottinghamDatabase",
    "RemoteABCFolderDataset",
    "RemoteDataset",
    "RemoteFolderDataset",
    "RemoteMusicDataset",
    "WikifoniaDataset",
    "get_dataset",
    "list_datasets",
]
